# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    auth0_domain: str = "coldquanta.auth0.com"
    auth0_client_id: str = "cTWqizVsJj2WrLtBgzgo3KZFgEY3k2OC"
    auth0_scope: str = "profile openid mail"
    auth0_audience: str = "https://oqtant.infleqtion.com/oqtant"
    auth0_notebook_redirect_uri: str = "http://localhost:8888/callback/"
    auth0_response_type: str = "token id_token"
    base_url: str = "https://oqtant.infleqtion.com/api"
    max_ind_var: int = 2
    run_list_limit: int = 30



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import warnings
from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from bert_schemas import job as job_schema
from bert_schemas import projected, timing
from bert_schemas.job import NonPlotOutput, PlotOutput
from bert_schemas.projected import Settings as ProjectedSettings
from ipyauth import Auth
from pydantic import BaseModel, confloat
from scipy.interpolate import interp1d

from oqtant import oqtant_client as oq
from oqtant.schemas.job import OqtantJob
from oqtant.schemas.optical import Barrier, Landscape, Laser, Pulse, Snapshot
from oqtant.schemas.output import OqtantNonPlotOutput, OqtantPlotOutput
from oqtant.schemas.rf import ConversionError, RfEvap, RfSequence, RfShield
from oqtant.simulator import Simulator
from oqtant.util import exceptions
from oqtant.util.auth import notebook_login

if TYPE_CHECKING:
    from oqtant.oqtant_client import OqtantClient

DEFAULT_RF_EVAP = RfEvap.new(
    times=[0, 50, 300, 800, 1100],
    powers=[600, 800, 600, 400, 400],
    frequencies=[21.12, 12.12, 5.12, 0.62, 0.02],
    interpolation=job_schema.InterpolationType.LINEAR,
)

DEFAULT_NAME = "quantum matter"
DEFAULT_LIFETIME = timing.EndTimeMs(10.0)  # ms
DEFAULT_TOF = timing.TimeOfFlightMs(12)  # ms
DEFAULT_IMAGE = job_schema.ImageType.TIME_OF_FLIGHT
TEMPERATURE_TO_EVAP_FREQUENCY = 0.067 / 200

pset = ProjectedSettings()


@dataclass
class OqtantLogin:
    access_token: str | None = None


class QuantumMatter(BaseModel):
    """A class that represents user inputs to create and manipulate quantum matter"""

    name: str | None = DEFAULT_NAME
    temperature: confloat(ge=0, le=500) | None = None
    lifetime: timing.EndTimeMs | None = DEFAULT_LIFETIME
    image: job_schema.ImageType | None = DEFAULT_IMAGE
    time_of_flight: timing.TimeOfFlightMs | None = DEFAULT_TOF
    rf_evap: RfEvap | None = None
    rf_shield: RfShield | None = None
    barriers: list[Barrier | job_schema.Barrier] | None = None
    landscape: Landscape | job_schema.OpticalLandscape | None = None
    lasers: list[Laser | job_schema.Laser] | None = None
    note: job_schema.JobNote | None = None
    client: object | None = None
    result: object | None = None
    job_id: str | None = None
    output: object | None = None
    is_sim: bool = False
    sim: Simulator | None = None
    run: int = 1

    def model_post_init(self, *args) -> None:
        if (self.temperature is not None) and (self.rf_evap is not None):
            warnings.warn(
                "Both 'temperature' and 'rf_evap' inputs provided, the last rf_evap frequency \
                will be altered."
            )
        end_time = 0.0
        lifetime = timing.decimal_to_float(self.lifetime)
        if self.barriers:
            for barrier in self.barriers:
                if isinstance(barrier, job_schema.Barrier):
                    barrier = Barrier.from_input(barrier)
                end_time = max(end_time, barrier.death)
            if end_time > lifetime:
                raise ValueError(
                    "Specified 'lifetime' not sufficient for constituent barrier object(s)."
                )
        if self.landscape:
            for snapshot in self.landscape.snapshots:
                if isinstance(snapshot, job_schema.Landscape):
                    snapshot = Snapshot.from_input(snapshot)
                end_time = max(end_time, timing.decimal_to_float(snapshot.time_ms))
            if end_time > lifetime:
                raise ValueError(
                    "Specified 'lifetime' not sufficient for constituent snapshot objects."
                )
        if self.lasers:
            for laser in self.lasers:
                if isinstance(laser, job_schema.Laser):
                    laser = Laser.from_input(laser)
                end_time = max(end_time, timing.decimal_to_float(max(laser.pulses[-1].times_ms)))
            if end_time > lifetime:
                raise ValueError(
                    "Specified 'lifetime' not sufficient for constituent laser pulse(s)."
                )

    @property
    def job_type(self) -> job_schema.JobType:
        """Property to get the job type of a submitted QuantumMatter object

        Returns:
            bert_schemas.job.JobType: The type of the job
        """
        if self.result:
            return self.result.job_type

    @property
    def status(self) -> job_schema.JobStatus:
        """Property to get the job status of a submitted QuantumMatter object

        Returns:
            bert_schemas.job.JobStatus: The status of the job
        """
        if self.result:
            return self.result.status

    @property
    def run_count(self) -> int:
        """Property to get the number of job runs for a submitted QuantumMatter object

        Returns:
            int: The total run number of the job
        """
        if self.result:
            return self.result.input_count

    @property
    def time_submit(self) -> int:
        """Property to get the time the current job was submitted

        Returns:
            str: The time the current job was submitted
        """
        if self.result:
            return self.result.time_submit

    @property
    def time_start(self) -> int:
        """Property to get the time the current job was run

        Returns:
            str: The time the current job was run
        """
        if self.result:
            return self.result.time_start

    @property
    def time_complete(self) -> int:
        """Property to get the time the current job was completed

        Returns:
            str: The time the current job was completed
        """
        if self.result:
            return self.result.time_complete

    @property
    def input(self) -> job_schema.InputValues:
        """Property to get the job input values of a QuantumMatter object

        Returns:
            bert_schemas.job.InputVales: The job input values of the QuantumMatter object's
            current result
        """
        return job_schema.InputValues(
            end_time_ms=self.lifetime,
            time_of_flight_ms=self.time_of_flight,
            image_type=self.image,
            rf_evaporation=self.rf_evaporation,
            optical_barriers=self.barriers,
            optical_landscape=self.landscape,
            lasers=self.lasers,
        )

    @property
    def rf_evaporation(self) -> job_schema.RfEvaporation:
        """Property to get the RF evaporation data of a QuantumMatter object

        Returns:
            bert_schemas.job.RfEvaporation: The RF evaporation values of the job's current result
        """
        rf_evap = deepcopy(self.rf_evap) if self.rf_evap is not None else DEFAULT_RF_EVAP
        if self.temperature is not None:
            rf_evap.frequencies_mhz[-1] = TEMPERATURE_TO_EVAP_FREQUENCY * self.temperature
        rf_evaporation = job_schema.RfEvaporation(
            times_ms=rf_evap.times_ms,
            frequencies_mhz=rf_evap.frequencies_mhz,
            powers_mw=rf_evap.powers_mw,
            interpolation=rf_evap.interpolation,
        )
        if self.rf_shield is not None:
            rf_evaporation.times_ms.append(self.lifetime)
            rf_evaporation.powers_mw.append(self.rf_shield.power)
            rf_evaporation.frequencies_mhz.append(self.rf_shield.frequency)

        return rf_evaporation

    @classmethod
    def from_input(
        cls,
        name: str,
        input: job_schema.InputValues,
        note: job_schema.JobNote | None = None,
        client: OqtantClient | None = None,
    ) -> QuantumMatter:
        """Method to create a new QuantumMatter object using the input values of an existing job

        Args:
            name (str): Name of the quantum matter
            input (bert_schemas.job.InputValues): The input values
            note (bert_schemas.job.JobNote | None): The notes for the input, can be None
            client (oqtant.oqtant_client.OqtantClient | None, optional): An instance of OqtantClient

        Returns:
            QuantumMatter: A new QuantumMatter object created using the input data
        """
        try:
            evap = RfEvap.from_input(input.rf_evaporation)
        except ConversionError:
            evap = None
        try:
            shield = RfShield.from_input(input.rf_evaporation)
        except ConversionError:
            shield = None

        barriers = None
        if input.optical_barriers:
            barriers = []
            for barrier in input.optical_barriers:
                barriers.append(Barrier.from_input(barrier))

        landscape = None
        if input.optical_landscape:
            landscape = Landscape.from_input(input.optical_landscape)

        return cls(
            name=name,
            lifetime=input.end_time_ms,
            image=input.image_type,
            time_of_flight=input.time_of_flight_ms,
            rf_evap=evap,
            rf_shield=shield,
            barriers=barriers,
            landscape=landscape,
            lasers=input.lasers,
            client=client,
            note=note,
        )

    @classmethod
    def from_oqtant_job(cls, job: OqtantJob, client: OqtantClient) -> QuantumMatter:
        """Method to create a new QuantumMatter object using an existing OqtantJob

        Args:
            job (oqtant.schemas.job.OqtantJob): The OqtantJob object to create from
            client (oqtant.oqtant_client.OqtantClient): An instance of OqtantClient
            run (int, optional): The specific run to use

        Returns:
            QuantumMatter: A new QuantumMatter object created using the OqtantJob data
        """
        inputs = job.inputs[0]
        qm = QuantumMatter.from_input(
            name=job.name, input=inputs.values, note=inputs.notes, client=client
        )
        qm.job_id = str(job.external_id)
        qm.result = job
        qm.run = job.run

        if inputs.output:
            qm.output = QuantumMatter.output_values_to_oqtant_output(inputs.output.values)
        return qm

    @staticmethod
    def output_values_to_oqtant_output(
        output_values: PlotOutput | NonPlotOutput,
    ) -> OqtantPlotOutput | OqtantNonPlotOutput:
        """Method to convert a completed job's output values to OqtantOutput

        Args:
            output_values (PlotOutput|NonPlotOutput):  The output values to convert
            to OqtantPlotOutput

        Returns:
            (oqtant.schemas.output.OqtantPlotOutput | oqtant.schemas.output.OqtantNonPlotOutput):
            The converted output values
        """

        if hasattr(output_values, "it_plot"):
            return OqtantPlotOutput(**output_values.model_dump())
        else:
            return OqtantNonPlotOutput(**output_values.model_dump())

    def write_to_file(self, *args, **kwargs) -> None:
        """Method to write the results of a submitted QuantumMatter object to a file.
        Wrapper for OqtantClient.write_job_to_file
        """
        self.client.write_job_to_file(self.result, *args, **kwargs)

    def submit_sim(self) -> None:
        """Method to submit a QuantumMatter object to be run as a simulation"""
        self.is_sim = True
        result = self.client.submit_sim(self)
        print("submitted simulation")
        self.sim = result

    def submit(self, track: bool = False, sim: bool = False) -> None:
        """Method to submit a QuantumMatter object to Oqtant to become a job and run on hardware or
        as a simulation

        Args:
            track (bool, optional): Flag to poll for job updates after submission
            sim (bool, optional): Flag to use the simulator backend instead of real hardware
        """
        if sim:
            self.submit_sim()
            return

        self.is_sim = False

        if self.job_id:
            raise exceptions.JobError(
                f"QuantumMatter object '{self.name}' was previously submitted and already has a job id.  "
                "Resubmission is not supported."
            )
        result = self.client.submit(self, track)
        self.job_id = result

    def get_sim_result(self) -> None:
        """Method to get the results of a simulator job. Alerts the user if simulation results are
        invalid due to boundary collision.
        """
        self.sim = self.client.get_sim_result(self.sim, self.image)

        output = None

        if self.image == job_schema.ImageType.IN_TRAP:
            it_output = self.sim.it_plot
            output = OqtantPlotOutput(**it_output)

        if self.image == job_schema.ImageType.TIME_OF_FLIGHT:
            tof_output = self.sim.tof_output
            output = OqtantNonPlotOutput(**tof_output)

        self.output = QuantumMatter.output_values_to_oqtant_output(output)

        self.result = self.client.convert_matter_to_job(self)
        if self.result.run is None:
            self.result.inputs[0].run = 1
            self.result.input_count = 1

        self.result.status = job_schema.JobStatus.COMPLETE
        job_done_msg = "Simulation complete."
        print(f"{job_done_msg:<20}")

        if self.sim.hit_boundary:
            print(
                "Results may be invalid after atoms reached simulation boundaries. "
                "Consider reducing barrier height or increasing ramp/evolution time."
            )

    def get_result(self, run: int | None = None) -> None:
        """Method to get the results of a hardware job

        Args:
            run (int, optional): The specific run to get
        """
        if self.is_sim:
            self.get_sim_result()
            return

        if not self.job_id:
            raise exceptions.JobError("Matter must be submitted before retrieving results")

        # use current run if one is not provided
        run = run if run else self.run
        self.result = self.client.get_job(self.job_id, run)

        # update matter input if current run is different than provided run
        if self.run != run:
            matter = self.from_input(
                name=self.name,
                input=self.result.inputs[0].values,
                note=self.result.inputs[0].notes,
                client=self.client,
            )
            self.lifetime = matter.lifetime
            self.image = matter.image
            self.time_of_flight = matter.time_of_flight
            self.rf_evap = matter.rf_evap
            self.rf_shield = matter.rf_shield
            self.barriers = matter.barriers
            self.lasers = matter.lasers
            self.note = matter.note
            self.run = run

        # clear output and update if new data exists
        self.output = None
        if self.result.status == job_schema.JobStatus.COMPLETE:
            self.output = QuantumMatter.output_values_to_oqtant_output(
                self.result.inputs[0].output.values
            )

    def corrected_rf_power(self, frequency_mhz: float, power_mw: float) -> float:
        """Method to calculate the corrected RF power based on the given frequency and power

        Args:
            frequency_mhz (float): The frequency in MHz
            power (float): The power in mW

        Returns:
            float: The corrected RF power in mW
        """
        voltage = power_mw * 5.0 / 1000.0  # payload power to attenuator voltage
        power_dbm = -26.2 - 42 * np.exp(-0.142 * frequency_mhz) - 32.76 * np.exp(-1.2 * voltage)

        # based on data taken from smallbert for power measured in dbm by a nearby pickup rf loop
        # as a function of the frequency (in MHz) and RF attenuator voltage (in volts)
        # the 'composer' turns payload powers of 0-1000 mW into voltages using a linear
        # relationship that maps 0-1000 mW to 0-5 V on the RF attenuator (5V = max power)

        return (1000.0 / 2.0e-3) * 10 ** (power_dbm / 10.0)  # dbm to mW with overall scaling

    def corrected_rf_powers(self, frequencies: list[float], powers: list[float]) -> list[float]:
        """Method to calculate the corrected RF powers based on the given lists of frequencies
        and powers

        Args:
            frequencies (list[float]): The frequencies in MHz
            powers (list[float]): The powers in mW

        Returns:
            list[float]: The corrected list of RF powers in mW
        """
        return [self.corrected_rf_power(freq, pow) for freq, pow in zip(frequencies, powers)]

    def show_rf_dynamics(self, corrected: bool = False) -> None:
        """Method to plot the dynamics of a QuantumMatter object's RF output

        Args:
            corrected (bool, optional): Flag to correct the RF power
        """
        evap = self.rf_evaporation
        rf_evap = RfEvap.from_input(evap)
        tstart = min(rf_evap.times_ms)
        evap_times = np.linspace(tstart, 0, num=int(abs(tstart) / 10), endpoint=True)
        fig, ax1 = plt.subplots()
        lns = []
        labs = []

        # plot of rf frequency vs time
        color = next(ax1._get_lines.prop_cycler)["color"]
        ax1.set_xlabel("time (ms)")
        ax1.set_ylabel("frequency (MHz)")
        ax1.set_ylim([0, 25])
        (ln1,) = plt.plot(evap_times, rf_evap.get_frequencies(evap_times), color=color)
        lns.append(ln1)
        labs.append("frequency")
        plt.plot(
            rf_evap.times_ms,
            rf_evap.get_frequencies(rf_evap.times_ms),
            ".",
            color=color,
        )
        if self.rf_shield is not None:
            plt.plot(
                [0, self.lifetime],
                [self.rf_shield.frequency] * 2,
                marker=".",
                color=color,
            )

        # plot of rf power vs time, on the same time axis as ax1
        ax2 = ax1.twinx()
        ax2.set_ylim([0, 1000])
        ax2.set_ylabel("power (mW)")
        color = next(ax1._get_lines.prop_cycler)["color"]
        (ln2,) = plt.plot(evap_times, rf_evap.get_powers(evap_times), color=color)
        lns.append(ln2)
        labs.append("power")
        plt.plot(
            rf_evap.times_ms,
            rf_evap.get_powers(rf_evap.times_ms),
            ".",
            color=color,
        )
        if self.rf_shield is not None:
            plt.plot([0, self.lifetime], [self.rf_shield.power] * 2, marker=".", color=color)
        if corrected:
            (ln3,) = plt.plot(
                evap_times,
                self.corrected_rf_powers(
                    rf_evap.get_frequencies(evap_times),
                    rf_evap.get_powers(evap_times),
                ),
                "--",
                color=color,
            )
            if self.rf_shield is not None:
                plt.plot(
                    [0, self.rf_shield.lifetime],
                    self.corrected_rf_powers(
                        [self.rf_shield.frequency] * 2,
                        [self.rf_shield.power] * 2,
                    ),
                    "--",
                    color=color,
                )
            lns.append(ln3)
            labs.append("corrected power")
        # shared setup
        ax1.legend(lns, labs, loc="upper center")
        color = next(ax1._get_lines.prop_cycler)["color"]
        plt.axvline(x=0, linestyle="dashed", color=color)
        plt.title("RF dynamic behavior")
        fig.tight_layout()  # avoid clipping right y-axis label
        plt.show()

    def get_magnetic_potential(self, positions: list[float]) -> list[float]:
        """Method to calculate the magnetic potentials for a given set of positions

        Args:
            positions (list[float]): The positions at which to calculate the potentials

        Returns:
            list[float]: List of magnetic potentials in kHz corresponding to the given positions
        """
        w = 2 * np.pi * 50  # weak axis trap frequency
        m = 87 * 1.66054e-27
        h = 6.626e-34

        # U = mf * g * ub * |B| with B = B0 + 0.5 * m * w^2 * x^2
        # for this purpose, we will set B0 = 0
        # (magnetic potential referenced to trap bottom as rf frequencies are)
        # our measured trap frequency is ~ 50 Hz

        potentials = 0.5 * m * w**2 * np.square(1e-6 * np.asarray(positions))  # in J
        potentials_khz = potentials / h / 1000.0  # in kHz
        return list(potentials_khz)

    def get_ideal_optical_potential(self, time: float, positions: list[float]) -> list[float]:
        """Method to calculate the "ideal" optical potential from constituent optical objects

        Args:
            time (float): time, in ms, for which the optical potential should be evaluated
            positions (list[float]): positions, in microns, where potential should be evaluated

        Returns:
            list[float]: list of potential energies, in kHz, at the request time and positions
        """
        potential = np.zeros_like(positions)
        if self.barriers is not None:
            for barr in self.barriers:
                potential += np.asarray(barr.get_ideal_potential(time=time, positions=positions))
        if self.landscape is not None:
            potential += np.asarray(
                self.landscape.get_ideal_potential(time=time, positions=positions)
            )
        return list(potential)

    def get_potential(
        self, time: float, positions: list[float], include_magnetic: bool = True
    ) -> list[float]:
        """Method to calculate the optical and magnetic potential at the given time for each
        position

        Args:
            time (float): The time at which to calculate the potential
            positions (list[float]): The positions at which to calculate the potential
            include_magnetic (bool, optional): Flag to include contributions from magnetic trap
        Returns:
            list[float]: List of potential energy corresponding to each request position
        """
        potential = np.asarray(
            projected.get_actual_potential(
                self.get_ideal_optical_potential, time=time, positions=positions
            )
        )
        if include_magnetic:
            potential += np.asarray(self.get_magnetic_potential(positions=positions))
        return list(potential)

    def show_potential(
        self,
        times: list[float] = [0.0],
        xlimits: list[float] = [pset.min_position - 1, pset.max_position + 1],
        ylimits: list[float] = [0.0 - 1, pset.max_energy + 1],
        include_ideal: bool = False,
        include_magnetic: bool = True,
        include_terminator: bool = False,
    ) -> None:
        """Method to plot the (optical) potential energy surface at the specified times

        Args:
            times (list[float], optional): The times for which to display the potential energy
            xlimits (list[float], optional): The plot limits for the x axis
            ylimits (list[float], optional): The plot limits for the y axis
            include_ideal (bool, optional): Flag for including target potential in plot
            include_magnetic (bool, optional): Flag to include contributions from magnetic trap
            include_terminator (bool, optional): Flag to include the position of the terminator beam
            relative to the trap
        """

        GAUSSIAN_1e2_DIAM_TO_FULL_WIDTH_HALF_MAX = 1.18
        term_radius_1e2 = pset.terminator_width
        full_width_half_max = GAUSSIAN_1e2_DIAM_TO_FULL_WIDTH_HALF_MAX * term_radius_1e2

        positions = np.arange(pset.min_position, pset.max_position + 0.1, 0.1, dtype=float)
        fig, ax = plt.subplots()

        plotted_term_already = False
        lns = []
        labs = []
        ln_term = []
        labs_term = []

        for time in times:
            color = next(ax._get_lines.prop_cycler)["color"]
            (ln,) = plt.plot(
                positions,
                self.get_potential(time, positions, include_magnetic),
                color=color,
                label=" ",
            )
            lns.append(ln)
            labs.append("t = " + str(time) + " ms")
            if include_ideal:
                potential = np.asarray(
                    self.get_ideal_optical_potential(time=time, positions=positions)
                )
                if include_magnetic:
                    potential += np.asarray(self.get_magnetic_potential(positions=positions))
                (ln2,) = plt.plot(positions, potential, "--", color=color, label=" ")
                lns.append(ln2)
                labs.append("t = " + str(time) + " ms (ideal)")

            if (
                self.lasers
                and self.lasers[0].type == job_schema.LaserType.TERMINATOR
                and not plotted_term_already
            ):
                term_center = self.lasers[0].position_um
                if include_terminator and self.lasers[0].is_on(time):
                    plt.vlines([term_center], ylimits[0], ylimits[1], "r", "--")
                    x_fwhm = [
                        term_center - full_width_half_max / 2,
                        term_center + full_width_half_max / 2,
                    ]
                    x_1e2 = [
                        term_center - term_radius_1e2,
                        term_center + term_radius_1e2,
                    ]
                    plt.fill_between(
                        x_fwhm,
                        [ylimits[0], ylimits[0]],
                        [ylimits[1], ylimits[1]],
                        alpha=0.5,
                        color="r",
                    )
                    (ln,) = plt.fill(np.NaN, np.NaN, "r", alpha=0.5, label=" ")
                    ln_term.append(ln)
                    labs_term.append("FWHM ($\\mu$m)")
                    plt.fill_between(
                        x_1e2,
                        [ylimits[0], ylimits[0]],
                        [ylimits[1], ylimits[1]],
                        alpha=0.1,
                        color="r",
                    )
                    (ln,) = plt.fill(np.NaN, np.NaN, "r", alpha=0.1, label=" ")
                    ln_term.append(ln)
                    labs_term.append("gaussian width (1/$e^2$) ($\\mu$m)")

                    plotted_term_already = True

        plt.xlabel("position (microns)", labelpad=6)
        plt.ylabel("potential energy (kHz)", labelpad=6)
        plt.xlim(xlimits)
        plt.ylim(ylimits)
        plt.legend()

        l2 = ax.legend(lns, labs, loc=2)
        if include_terminator:
            ax.legend(ln_term, labs_term, loc=1)
            plt.gca().add_artist(l2)
        plt.show()

    def show_barrier_dynamics(self) -> None:
        """Method to plot the time dynamics of every Barrier object within a QuantumMatter object"""
        fig, (ax1, ax2, ax3) = plt.subplots(nrows=3, ncols=1, sharex=True, figsize=(6, 6))
        fig.suptitle("Barrier dynamics")
        ax1.set_xlim([-1, self.input.end_time_ms])
        ax1.set_ylabel("position (microns)")
        ax2.set_ylabel("height (kHz)")
        ax3.set_ylabel("width (microns)")
        ax3.set_xlabel("time (ms)")
        lns = []
        labs = []

        style = "steps-pre"
        for indx, barrier in enumerate(self.barriers):
            color = next(ax1._get_lines.prop_cycler)["color"]
            tstart = timing.decimal_to_float(min(barrier.times_ms))
            tstop = timing.decimal_to_float(max(barrier.times_ms))
            times = np.linspace(tstart, tstop, num=int((tstop - tstart) / 0.1), endpoint=True)
            barrier_times = timing.decimals_to_floats(barrier.times_ms)
            (ln,) = ax1.plot(times, barrier.get_positions(times), color=color, drawstyle=style)
            ax1.plot(barrier_times, barrier.get_positions(barrier_times), ".", color=color)
            ax2.plot(times, barrier.get_heights(times), color=color, drawstyle=style)
            ax2.plot(
                barrier_times,
                barrier.get_heights(barrier_times),
                ".",
                color=color,
            )
            ax3.plot(times, barrier.get_widths(times), color=color, drawstyle=style)
            ax3.plot(barrier_times, barrier.get_widths(barrier_times), ".", color=color)
            lns.append(ln)
            labs.append("barrier " + str(indx + 1))
        fig.legend(lns, labs)
        plt.show()

    def show_laser_pulse_timing(self, figsize=(6, 4)) -> None:
        """
        Method to plot the timing of a single terminator pulse in the experiment

        Args:
            figsize (tuple, optional) : Size of the output plot

        """

        fig, ax1 = plt.subplots(figsize=figsize)
        ax2 = ax1.twinx()
        lns = []
        labs = []

        # TODO support plotting for more than one laser, more than one pulse per laser

        if self.lasers:
            for laser in self.lasers:
                # sample on 50us time steps
                # TODO: support calculation and plotting of non-square pulses
                times = np.arange(0, timing.decimal_to_float(self.lifetime), 50e-3)
                intensities = np.zeros(len(times))
                detunings = np.zeros(len(times))

                # add all the pulse times, detunings, and intensities
                for pulse in laser.pulses:
                    interpolation = job_schema.interpolation_to_kind(pulse.interpolation)
                    f_pulse = interp1d(
                        timing.decimals_to_floats(pulse.times_ms),
                        pulse.intensities_mw_per_cm2,
                        kind=interpolation,
                        bounds_error=False,
                        fill_value=0,
                    )
                    intensities_pulse = f_pulse(times)
                    intensities += intensities_pulse
                    det_idx = np.argwhere(
                        (times >= timing.decimal_to_float(pulse.times_ms[0]))
                        & (times <= timing.decimal_to_float(pulse.times_ms[-1]))
                    )
                    detunings[det_idx] = pulse.detuning_mhz

                (ln,) = ax1.plot(times, intensities, "g")
                labs.append("intensities")
                lns.append(ln)
                (ln,) = ax2.plot(times, detunings, "b")
                lns.append(ln)
                labs.append("detunings ")

        ax1.set_xlabel("Time (ms)")
        ax1.set_ylabel("Intensity (mW/$cm^2$)", color="g")
        ax2.set_ylabel("Detuning (MHz)", color="b")

        fig.legend(lns, labs, bbox_to_anchor=(0.8, 0.3))
        fig.suptitle("Laser pulse timing during in-trap experiment")
        plt.show()

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True


class QuantumMatterFactory:
    """An abstract factory for creating instances of the QuantumMatter schema classes"""

    def __init__(self):
        self.login = notebook_login()
        self.client = None

    def get_login(self) -> Auth:
        """Method to display the authentication widget inside of a notebook, if no access token file
        is found.

        Returns:
            ipyauth.Auth: The authentication widget
        """
        access_token = None
        try:
            with open(os.path.join(os.path.expanduser("~"), ".access-token.txt")) as file:
                access_token = file.read()
        except:  # noqa: E722
            ...
        if access_token:
            self.login = OqtantLogin(access_token=access_token)
        else:
            return self.login

    def get_client(self, token: str | None = None) -> None:
        """Method to get an instance of OqtantClient and assign it to self.client

        Args:
            token (str | None, optional): Token to use when working outside of a notebook
        """
        access_token = token or self.login.access_token
        self.client = oq.get_oqtant_client(access_token)

    def search_jobs(self, *args, **kwargs) -> list[dict]:
        """Method to search for jobs.
        Wrapper for OqtantClient.search_jobs

        Returns:
            list[dict]: The jobs found for the search criteria
        """
        return self.client.search_jobs(*args, **kwargs)

    def show_queue_status(self, *args, **kwargs) -> list[dict]:
        """Method to show the current queue status of jobs submitted by the authenticated user.
        Wrapper for OqtantClient.show_queue_status

        Returns:
            list[dict]: The jobs found for the search criteria along with their queue status
        """
        return self.client.show_queue_status(*args, **kwargs)

    def show_job_limits(self) -> dict:
        """Method to show the current job limits of the authenticated user.
        Wrapper for OqtantClient.show_job_limits

        Returns:
            dict: The job limit information for the authenticated user
        """
        return self.client.show_job_limits()

    def load_matter_from_file(self, *args, **kwargs) -> QuantumMatter:
        """Method to create a QuantumMatter object using data in a file.
        Wrapper for OqtantClient.load_job_from_file

        Returns:
            QuantumMatter: A new QuantumMatter object created using the file data
        """
        job = self.client.load_job_from_file(*args, **kwargs)
        return QuantumMatter.from_oqtant_job(job, self.client)

    def load_matter_from_job_id(self, job_id: str, run: int = 1) -> QuantumMatter:
        """Method to create a QuantumMatter object using data from an existing job in the database

        Args:
            job_id (str): The id of the job to get from the database
            run (int, optional): The specific run to get

        Returns:
            QuantumMatter: A new QuantumMatter object created using the jobs data
        """
        result = self.client.get_job(job_id, run)
        if result.status == job_schema.JobStatus.COMPLETE:
            output_values = result.inputs[0].output.values
            output = QuantumMatter.output_values_to_oqtant_output(output_values)
        else:
            output = None

        matter = self.create_quantum_matter_from_input(
            name=result.name, input=result.inputs[0].values, note=result.inputs[0].notes
        )
        matter.output = output
        matter.job_id = job_id
        matter.result = result
        matter.run = run
        return matter

    def submit_list_as_batch(self, *args, **kwargs) -> str:
        """Method to submit multiple QuantumMatter objects as a single job.
        Wrapper for OqtantClient.submit_list_as_batch

        Returns:
            str: The ID of the submitted job
        """
        return self.client.submit_list_as_batch(*args, **kwargs)

    def create_quantum_matter(
        self,
        name: str | None = None,
        temperature: float | None = None,
        lifetime: float | None = None,
        image: job_schema.ImageType | None = None,
        time_of_flight: float | None = None,
        rf_evap: RfEvap | None = None,
        rf_shield: RfShield | None = None,
        barriers: list[Barrier] | None = None,
        landscape: Landscape | None = None,
        lasers: list[job_schema.Laser] | None = None,
        note: str | None = None,
    ) -> QuantumMatter:
        """Method to create a QuantumMatter object

        Args:
            name (str | None, optional): The name of the quantum matter
            temperature (float | None, optional): The quantum matter temperature
            lifetime (float | None, optional): The quantum matter lifetime
            image (bert_schemas.job.ImageType | None, optional): The quantum matter image type
            time_of_flight (float | None, optional): The quantum matter time of flight
            rf_evap (oqtant.schemas.rf.RfEvap | None, optional): The quantum matter RF evaporation
            rf_shield (oqtant.schemas.rf.RfShield | None, optional): The quantum matter RF shield
            barriers (list[oqtant.schemas.optical.Barrier] | None, optional): The quantum matter barriers
            landscape (oqtant.schemas.optical.Landscape | None, optional): The quantum matter landscape
            lasers (list[bert_schemas.job.Lasers] | None, optional): The quantum matter lasers
            note (str | None, optional): A note about the quantum matter

        Returns:
            QuantumMatter: A new QuantumMatter object
        """
        kwargs = {"client": self.client}
        for k, v in locals().items():
            if v is not None:
                kwargs[k] = v

        return QuantumMatter(**kwargs)

    def create_quantum_matter_from_input(
        self,
        name: str,
        input: job_schema.InputValues,
        note: job_schema.JobNote | None = None,
    ) -> QuantumMatter:
        """Method to create a QuantumMatter object using the input values of a job.
        Wrapper for QuantumMatter.from_input

        Args:
            name (str): The name of the quantum matter
            input (bert_schemas.job.InputValues): The input values
            note (job_schema.job.JobNote | None): The notes for the input, can be None

        Returns:
            QuantumMatter: A new QuantumMatter object created using the input data
        """
        return QuantumMatter.from_input(name, input, note, self.client)

    @staticmethod
    def create_snapshot(
        time: float = 0,
        positions: list = [-10, 10],
        potentials: list = [0, 0],
        interpolation: job_schema.InterpolationType = "LINEAR",
    ) -> Snapshot:
        """Method to create a Snapshot object

        Args:
            time (float, optional): The time in milliseconds
            positions (list, optional): A list of positions in micrometers
            potentials (list, optional): A list of potentials in kilohertz
            interpolation (bert_schemas.job.InterpolationType, optional): The type of interpolation for spatial data

        Returns:
            oqtant.schemas.optical.Snapshot: A new Snapshot object
        """
        return Snapshot(
            time_ms=time,
            positions_um=positions,
            potentials_khz=potentials,
            spatial_interpolation=interpolation,
        )

    @staticmethod
    def create_snapshot_from_input(input: job_schema.Landscape) -> Snapshot:
        """Method to create a Snapshot object using the input values of a job

        Args:
            input (bert_schemas.job.Landscape): The landscape input data

        Returns:
            oqtant.schemas.optical.Snapshot: A new Snapshot object created using the input data
        """
        return Snapshot(**input.model_dump())

    @staticmethod
    def create_landscape(
        snapshots: list[Snapshot] = [Snapshot.new(time=0), Snapshot.new(time=2)]
    ) -> Landscape:
        """Method to create a Landscape object from a list Snapshot objects

        Args:
            snapshots (list[oqtant.schemas.optical.Snapshot], optional): List of snapshots, defaults if not provided

        Returns:
            oqtant.schemas.optical.Landscape: A new Landscape object
        """
        optical_landscapes = []
        for snapshot in snapshots:
            optical_landscapes.append(  # kludge!
                job_schema.Landscape(
                    time_ms=snapshot.time_ms,
                    positions_um=snapshot.positions_um,
                    potentials_khz=snapshot.potentials_khz,
                    spatial_interpolation=snapshot.spatial_interpolation,
                )
            )
        return Landscape(landscapes=optical_landscapes)

    @staticmethod
    def create_landscape_from_input(input: job_schema.OpticalLandscape) -> Landscape:
        """Method to create a Landscape object from the input values of a job

        Args:
            input (bert_schemas.job.OpticalLandscape): The input values

        Returns:
            oqtant.schemas.optical.Landscape: A new Landscape object created using the input data
        """
        return Landscape(**input.model_dump())

    @staticmethod
    def create_barrier(
        positions: list[float] = [0.0, 0.0],
        heights: list[float] = [0.0, 0.0],
        widths: list[float] = [1.0, 1.0],
        times: list[float] = [0.0, 10.0],
        shape: job_schema.ShapeType = "GAUSSIAN",
        interpolation: job_schema.InterpolationType = "LINEAR",
    ) -> Barrier:
        """Method to create a Barrier object

        Args:
            positions (list[float], optional): The barrier positions
            heights (list[float], optional): The barrier heights
            widths (list[float], optional): The barrier widths
            times (list[float], optional): The barrier times
            shape (bert_schemas.job.ShapeType, optional): The barrier shape
            interpolation (bert_schemas.job.InterpolationType, optional): The barrier interpolation type

        Returns:
            oqtant.schemas.optical.Barrier: A new Barrier object

        Raises:
            ValueError: if data lists are not of equal length
        """
        if not (len(positions) == len(heights) == len(widths) == len(times)):
            raise ValueError("Barrier data lists must be of equal length, default minimum of 2")

        data = {
            "times_ms": times,
            "positions_um": positions,
            "heights_khz": heights,
            "widths_um": widths,
            "shape": shape,
            "interpolation": interpolation,
        }

        return Barrier(**data)

    @staticmethod
    def create_barrier_from_input(input: job_schema.Barrier) -> Barrier:
        """Method to create a Barrier object from the input values of a job

        Args:
            input (bert_schemas.job.Barrier): The input values

        Returns:
            oqtant.schemas.optical.Barrier: A new Barrier object created using the input data
        """
        return Barrier(**input.model_dump())

    @staticmethod
    def create_terminator(time_on: float, time_off: float) -> Laser:
        """Method to create terminator pulse

        Args:
            time_on (float) : time to start the pulse in ms
            time_off (float) : time to end the pulse in ms

        Returns:
            oqtant.schemas.optical.Laser: A new Laser object
        """

        pulse_data = {
            "times_ms": [time_on, time_off],
            "intensities_mw_per_cm2": [1, 1],
            "detuning_mhz": 0,
            "interpolation": job_schema.InterpolationType.OFF,
        }

        pulse = Pulse(**pulse_data)

        laser_data = {
            "type": job_schema.LaserType.TERMINATOR,
            "position_um": pset.terminator_position,
            "pulses": [pulse],
        }

        return Laser(**laser_data)

    @staticmethod
    def create_terminator_from_input(input: job_schema.Laser) -> Laser:
        """Method to create a Laser object from the input values of a job

        Args:
            input (bert_schemas.job.Laser): The input values

        Returns:
            oqtant.schemas.optical.Laser: A new Laser object created using the input data
        """
        return Laser(**input.model_dump())

    @staticmethod
    def create_rf_sequence(
        times: list = [0],
        powers: list = [0],
        frequencies: list = [0],
        interpolation: str = "LINEAR",
    ) -> RfSequence:
        """Method to create a RfSequence object

        Args:
            times (list[int], optional): The time values in milliseconds
            powers (list[list[float], optional): The power values in milliwatts
            frequencies (list[float], optional): The frequency values in megahertz
            interpolation (bert_schemas.job.InterpolationType, optional): The interpolation type to be used

        Returns:
            oqtant.schemas.rf.RfSequence: A new RfSequence object
        """
        return RfSequence(
            times_ms=times,
            powers_mw=powers,
            frequencies_mhz=frequencies,
            interpolation=interpolation,
        )

    @staticmethod
    def create_rf_sequence_from_input(input: job_schema.RfEvaporation) -> RfSequence:
        """Method to create a RfSequence object from the input values of a job

        Args:
            input (bert_schemas.job.RfEvaporation: The input values

        Returns:
            oqtant.schemas.rf.RfSequence: A new RfSequence object created using the input data
        """
        return RfSequence(**input.model_dump())

    @staticmethod
    def create_rf_evap(
        times: list = [0],
        powers: list = [0],
        frequencies: list = [0],
        interpolation: str = "LINEAR",
    ) -> RfEvap:
        """Method to create a RfEvap object

        Args:
            times (list[int], optional): The time values in milliseconds
            powers (list[list[float], optional): The power values in milliwatts
            frequencies (list[float], optional): The frequency values in megahertz
            interpolation (bert_schemas.job.InterpolationType, optional): The interpolation type to be used

        Returns:
            oqtant.schemas.rf.RfEvap: A new RfEvap object
        """
        return RfEvap.new(
            times=[t - max(times) for t in times],
            powers=powers,
            frequencies=frequencies,
            interpolation=interpolation,
        )

    @staticmethod
    def create_rf_evap_from_input(input: job_schema.RfEvaporation) -> RfEvap:
        """Method to create a RfEvap object from the input values of a job

        Args:
            input (bert_schemas.job.RfEvaporation): The input values

        Returns:
            oqtant.schemas.rf.RfEvap: A new RfEvap object created using the input data
        """
        return RfEvap.from_input(input)

    @staticmethod
    def create_rf_shield(
        power: float = 0, frequency: float = 0, lifetime: timing.EndTimeMs = 1.0
    ) -> RfShield:
        """Method to create a RfShield object

        Args:
            power (float, optional): The RfShield power
            frequency (float, optional): The RfShield frequency
            lifetime (float, optional): The RfShield lifetime

        Returns:
            oqtant.schemas.rf.RfShield: A new RfShield object
        """
        return RfShield.new(
            lifetime,
            frequency,
            power,
            interpolation="OFF",
        )

    @staticmethod
    def create_rf_shield_from_input(input: job_schema.RfEvaporation) -> RfShield:
        """Method to create a RfShield object from the input values of a job

        Args:
            input (bert_schemas.job.RfEvaporation): The input values

        Returns:
            oqtant.schemas.rf.RfShield: A new RfShield object created using the input data
        """
        return RfShield.from_input(input)



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from dataclasses import dataclass
from logging import getLogger
from typing import Final

import matplotlib.pyplot as plt
import numpy as np
from bert_schemas.job import Image, LineChart
from matplotlib.animation import FuncAnimation
from scipy import interpolate
from scipy.interpolate import RectBivariateSpline

from oqtant.simulator.qm_potential import QMPotential
from oqtant.simulator.wave_function import WaveFunction
from oqtant.util.exceptions import SimValueError

logger = getLogger(__name__)


@dataclass
class TimeSpan:
    start: float
    end: float


class Simulator:
    """
    'Simulator' Defines methods for evolution and plotting of the system described by the Oqtant simulator.
    The Oqtant simulator is constantly in evolution and the inteface should not be relied upon for use external to
    the Oqtant API.
    """

    number_of_atoms: Final[int] = 1e4
    # this is specific to interactions between Rubidium - 87 atoms
    scattering_length: Final[float] = 0.0034117
    interaction_strength: Final[float] = (
        4.0 * np.pi * number_of_atoms * scattering_length
    )
    EXPERIMENT_IT_TOF_OD_RATIO: Final[int] = 0.04057

    def __init__(self, potential: QMPotential):
        """
        Creates Three Wavefunctions:
        - One is In-Trap (IT)
        - One is Time of Flight Far Field (tof_ff)
        - One is Time of Flight Near Field (tof_nf)

        Args:
        potential: Defined with QMPotential object.
        """
        # No need for TOF version of self.qm_potential as it's switched off in this mode.
        self.qm_potential = potential
        self.wavefunction = WaveFunction()
        self.wavefunction_tof_nf = WaveFunction(tof_nf=True)
        self.wavefunction_tof_ff = WaveFunction(tof_ff=True)

        self.times: list | None = None
        self.psi_history = []  # None
        self.time_step = 0.5e-3
        self.sampling_rate = 100

        # use near-field grid for up to 10ms
        self.max_NF_time = 6 * self.qm_potential.oqt_time_to_sim

        # tof_nf mode uses larger grids and turns off the trap
        self.tof_nf = False

        # tof_ff uses even larger grids and turns off the trap
        self.tof_ff = False

        self.ground = False  # ground state mode. evolves in imaginary time
        self.clip = True  # clip:  should we restrict the maximum energy in the problem for stability purposes?
        self.last_it_time_idx = 0
        self.hit_boundary = False

    def set_ground_state(self) -> None:
        """
        This function evolves the condensate with ground = True.
        It is done at negative times before the barriers are switched on.
        2.5 simulation units of time is sufficient to settle down to the ground state.

        Returns:
            None

        """

        self.wavefunction.psi = self.wavefunction.initial_psi(
            sigma_r=1 / np.sqrt(self.qm_potential.w0r)
        )
        self.times = []
        self.psi_history = []
        self.ground = True  # we are in ground state mode
        self.run_rk4(TimeSpan(-2.5, -0.01))
        self.times = [0]  # reset times
        self.psi_history = [
            self.wavefunction.psi
        ]  # reset psi_history to begin from the ground state.

    def run_evolution(self) -> None:
        """
        Run the simulation in In-Trap (IT) mode. This function evolves the condensate with ground = False.
        It is done at positive times during the "experiment" stage (t=0).
        Optical potentials may be applied during this stage.
        It runs for the lifetime of the quantum_matter object
        It starts from the end result of running get_ground_state

        Returns:
            None

        """
        if self.times is None:
            raise ValueError("Ground State not set")
        self.ground = False  # we are no longer in the ground state mode

        # start from the end result of running get_ground_state
        # load the ground state found in set_ground_state
        self.wavefunction.psi = self.psi_history[0]
        if (
            self.qm_potential.lifetime > 0
        ):  # if lifetime is greater than zero, then we must evolve the wave function.
            self.psi_history = []  # clear psi_history
            self.times = []  # clear times
            self.run_rk4(TimeSpan(0, self.qm_potential.lifetime), stage_name="in-trap ")
        self.last_it_time_idx = len(self.times) - 1

    def run_TOF(self) -> None:
        """
        This function runs the TOF evolution and turns on/off the far field grid as needed

        Returns:
            None

        """

        self.run_TOF_nearField()
        # run in far field mode if user specifies long TOF
        if self.qm_potential.time_of_flight > self.max_NF_time:
            self.run_TOF_farField()

    def run_TOF_nearField(self) -> None:  # TOF mode
        """
        This function evolves the condensate with ground = False, with no trapping potentials or optical potentials.
        It runs for up to 6ms of the quantum_matter object, beginning with the result wavefunction of run_evolution
        It uses the near field TOF grid: self.wavefunction_tof_nf.three_d_grid is created with argument NF = True

        Returns:
            None

        """
        # we are now in "TOF mode".  Functions will use the TOF grids.
        self.tof_nf = True
        self.tof_ff = False  # we are not yet in far field mode.

        # load the final state from the IT evolution and interpolate it onto the near-field grids.
        # basically this loads wavefunction_tof
        self.wavefunction_tof_nf.psi = self.convert_intrap_to_nearfield(
            self.psi_history[-1]
        )

        # check if we need to go limit time in near-field mode.
        if self.qm_potential.time_of_flight > self.max_NF_time:
            end_nf_time = self.max_NF_time
        else:
            end_nf_time = self.qm_potential.time_of_flight

        # evolve using the near-field grids.
        self.run_rk4(
            TimeSpan(
                self.qm_potential.lifetime,
                self.qm_potential.lifetime + end_nf_time,
            ),
            stage_name="near field TOF ",
        )

    def run_TOF_farField(self) -> None:  # TOF mode
        """
        This function evolves the condensate with ground = False, with no trapping potentials or optical potentials.
        It runs for the remaining time of flight of the quantum_matter object, beginning with the result wavefunction
        of run_TOF_nearField at 6ms.
        It uses the far field TOF grid: self.wavefunction_tof_ff.three_d_grid is created with argument FF = True

        Returns:
            None

        """

        self.tof_nf = False  # we are no longer in near field mode.
        self.tof_ff = True  # we are now in far field mode.

        # load the final state from the NF evolution and interpolate it onto the three_d_grid_TOF grids.
        # basically this loads wavefunction_tof
        self.wavefunction_tof_ff.psi = self.convert_nearfield_to_farfield(
            self.psi_history[-1]
        )

        # evolve using far-field grids.
        self.run_rk4(
            TimeSpan(
                self.qm_potential.lifetime + self.max_NF_time,
                self.qm_potential.lifetime + self.qm_potential.time_of_flight,
            ),
            stage_name="far field TOF ",
        )

        return

    def convert_intrap_to_nearfield(self, psi: np.ndarray) -> np.ndarray:
        """
        Performs interpolation of wave function between IT and TOF grids
        For handoff between evolve IT and in TOF modes.

        Args:
            psi (ndarray):  a wave function of the IT grid (self.wavefunction.three_d_grid)
        Returns:
            ndarray:  psi interpolated onto the TOF nearfield grid (self.wavefunction_tof_nf.three_d_grid)
        """

        # construct an interpolation function for psi from the IT grid.  separate out into real and imag parts.
        ip_real = RectBivariateSpline(
            self.wavefunction.three_d_grid.r[:, 0],
            self.wavefunction.three_d_grid.x[0, :],
            np.real(psi),
            kx=3,
            ky=3,
        )  # cubic spline interpolation
        ip_imag = RectBivariateSpline(
            self.wavefunction.three_d_grid.r[:, 0],
            self.wavefunction.three_d_grid.x[0, :],
            np.imag(psi),
            kx=3,
            ky=3,
        )  # cubic spline interpolation

        # interpolate psi onto TOF grid
        psi_tof_nf = ip_real(
            self.wavefunction_tof_nf.three_d_grid.r[:, 0],
            self.wavefunction_tof_nf.three_d_grid.x[0, :],
        ) + 1j * ip_imag(
            self.wavefunction_tof_nf.three_d_grid.r[:, 0],
            self.wavefunction_tof_nf.three_d_grid.x[0, :],
        )

        return psi_tof_nf

    def convert_nearfield_to_farfield(self, psi_tof_nf: np.ndarray) -> np.ndarray:
        """
        Interpolation of wave function between NF and FF grids
        For handoff between evolve NF and in FF modes.

        Args:
            psi_tof_nf (ndarray):  a wave function of the NF grid (self.wavefunction_tof_nf.three_d_grid)
        Returns:
            ndarray:  psi interpolated onto the TOF grid (self.wavefunction_tof_ff.three_d_grid)
        """

        # construct an interpolation function for psi from the IT grid.  separate out into real and imag parts.
        ip_real = RectBivariateSpline(
            self.wavefunction_tof_nf.three_d_grid.r[:, 0],
            self.wavefunction_tof_nf.three_d_grid.x[0, :],
            np.real(psi_tof_nf),
            kx=3,
            ky=3,
        )  # cubic spline interpolation
        ip_imag = RectBivariateSpline(
            self.wavefunction_tof_nf.three_d_grid.r[:, 0],
            self.wavefunction_tof_nf.three_d_grid.x[0, :],
            np.imag(psi_tof_nf),
            kx=3,
            ky=3,
        )  # cubic spline interpolation

        # interpolate psi onto TOF grid
        psi_tof_ff = ip_real(
            self.wavefunction_tof_ff.three_d_grid.r[:, 0],
            self.wavefunction_tof_ff.three_d_grid.x[0, :],
        ) + 1j * ip_imag(
            self.wavefunction_tof_ff.three_d_grid.r[:, 0],
            self.wavefunction_tof_ff.three_d_grid.x[0, :],
        )

        return psi_tof_ff

    def get_laplacian(self, y: np.ndarray) -> np.ndarray:
        """
        Implementation of the second derivatives in x and r including forward, central, and backward formulas
        to second order accuracy

        Args:
            y (ndarray): function for which we calculate the laplacian

        Returns:
            ndarray: The laplacian of the function

        """
        if self.tof_nf:
            dr, dx = (
                self.wavefunction_tof_nf.three_d_grid.dr,
                self.wavefunction_tof_nf.three_d_grid.dx,
            )
        elif self.tof_ff:
            dr, dx = (
                self.wavefunction_tof_ff.three_d_grid.dr,
                self.wavefunction_tof_ff.three_d_grid.dx,
            )
        else:
            dr, dx = (
                self.wavefunction.three_d_grid.dr,
                self.wavefunction.three_d_grid.dx,
            )

        # First derivative.
        dydr = np.gradient(y, axis=0, edge_order=2) / dr
        dydr[0, :] = 0  # Enforce Neumann boundary condition

        #         # Central difference
        d2ydr2 = (np.roll(y, -1, axis=0) - 2 * y + np.roll(y, 1, axis=0)) / dr**2
        d2ydx2 = (np.roll(y, -1, axis=1) - 2 * y + np.roll(y, 1, axis=1)) / dx**2

        #         # Forward difference at start of array
        d2ydr2[0, :] = (2 * y[0, :] - 5 * y[1, :] + 4 * y[2, :] - y[3, :]) / dr**2
        d2ydx2[:, 0] = (2 * y[:, 0] - 5 * y[:, 1] + 4 * y[:, 2] - y[:, 3]) / dx**2

        #         # Backward difference at end of array
        d2ydr2[-1, :] = (
            -y[-4, :] + 4 * y[-3, :] - 5 * y[-2, :] + 2 * y[-1, :]
        ) / dr**2
        d2ydx2[:, -1] = (
            -y[:, -4] + 4 * y[:, -3] - 5 * y[:, -2] + 2 * y[:, -1]
        ) / dx**2

        if self.tof_nf:  # trying to avoid having to make new r grids
            # Laplacian in cylindrical coordinates with axial symmetry
            return (
                (1.0 / self.wavefunction_tof_nf.three_d_grid.r) * dydr + d2ydr2 + d2ydx2
            )

        elif self.tof_ff:
            return (
                (1.0 / self.wavefunction_tof_ff.three_d_grid.r) * dydr + d2ydr2 + d2ydx2
            )
        else:
            return (1.0 / self.wavefunction.three_d_grid.r) * dydr + d2ydr2 + d2ydx2

    def get_gpe(self, psi: np.ndarray) -> np.ndarray:
        """
        Implementation of the Gross-Pitaevskii Equation w/Neumann boundary conditions at r = 0
        and Dirichlet at large x and r.
        If self.tof_nf or self.tof_ff == True, the external potential is ignored.

        Args:
            psi (ndarray): the current timestep wavefunction

        Returns:
            ndarray: wavefunction calculated by the Gross-Pitaevskii Equation

        """
        # Enforce Neumann boundary condition
        psi[1, :] = psi[0, :]

        laplacian = self.get_laplacian(psi)

        # if we are not in TOF mode we are then in in-trap mode
        if self.tof_nf | self.tof_ff:
            return (0.5 * 1j) * laplacian - 1j * (
                self.interaction_strength * np.abs(psi) ** 2
            ) * psi
        else:
            return (0.5 * 1j) * laplacian - 1j * (
                self.qm_potential.potential
                + self.interaction_strength * np.abs(psi) ** 2
            ) * psi

    def run_rk4(self, time_span: TimeSpan, stage_name: str = "") -> None:
        """
        Implementation of the Runge-Kutta 4th order method to evolve in time.
        Depends on if the simulation is in IT mode, tof_nf or tof_ff mode.

        Args:
            time_span (TimeSpan):  a list of times (in milliseconds)
            stage_name (str): name for the evolving stage. default = ""
        Returns:
            None
        """
        # TODO: Refactor: ALB-6346
        if self.tof_nf:  # toggle between IT and TOF grids
            three_d_grid = self.wavefunction_tof_nf.three_d_grid
            wavefunction = self.wavefunction_tof_nf
        elif self.tof_ff:
            three_d_grid = self.wavefunction_tof_ff.three_d_grid
            wavefunction = self.wavefunction_tof_ff
        else:
            three_d_grid = self.wavefunction.three_d_grid
            wavefunction = self.wavefunction

        n = int(
            (time_span.end - time_span.start) / self.time_step
        )  # number of sub-intervals to evaluate rk4 on.
        t = np.linspace(
            time_span.start, time_span.end, n + 1
        )  # list of all time sub-intervals

        f = np.zeros(
            (three_d_grid.Nr, three_d_grid.Nx),
            dtype=complex,
        )  # stored wave function at each time

        # Set the initial wave function
        f = wavefunction.psi

        dt = self.time_step

        # if ground = True, we evolve in imaginary time
        if self.ground:
            dt = -dt * 1j

        oqt_times_us = t * QMPotential.sim_time_to_oqt / QMPotential.msec

        for i in range(n):
            # If in IT mode, we update the potential every 100 microseconds
            if not (self.tof_nf | self.tof_ff) and i == 0:
                self.qm_potential.update_potential(t[i], self.clip)

            if (
                not (self.tof_nf | self.tof_ff)
                and i > 0
                and (
                    oqt_times_us[i] % 100 < oqt_times_us[i - 1] % 100
                    and self.ground is False
                )
            ):
                self.qm_potential.update_potential(t[i], self.clip)

            psi = wavefunction.psi

            f1 = self.get_gpe(psi)
            f2 = self.get_gpe(psi + f1 * dt / 2)
            f3 = self.get_gpe(psi + f2 * dt / 2)
            f4 = self.get_gpe(psi + f3 * dt)

            f = f + dt * (f1 + 2.0 * f2 + 2.0 * f3 + f4) / 6.0

            # Dirichlet boundary conditions at asymptotes of simulation
            f[-1, :] = 0
            f[:, 0] = 0
            f[:, -1] = 0

            # Enforce Neumann boundary condition
            f[1, :] = f[0, :]

            wavefunction.psi = f
            wavefunction.psi = wavefunction.normalize(
                wavefunction.psi
            )  # Enforce normalization
            f = wavefunction.psi

            if i % self.sampling_rate == 0:
                self.psi_history.append(f)
                self.times.append(t[i])

                if not self.hit_boundary:
                    self.is_wavefunction_at_boundary(
                        f
                    )  # check if the condensate is hitting the simulation boundary.

                if not self.ground and len(self.times) > 1:
                    oqt_time = self.times[-1] * self.qm_potential.sim_time_to_oqt
                    total_time = (
                        self.qm_potential.lifetime + self.qm_potential.time_of_flight
                    ) * self.qm_potential.sim_time_to_oqt
                    print(
                        f"simulating: {stage_name:s} {oqt_time:.3f} of {total_time:.3f}ms         ",
                        end="\r",
                    )

    def is_wavefunction_at_boundary(self, psi: np.ndarray) -> None:
        """
        Warn user if the condensate is hitting the edge of the simulation.

        Args:
            psi (ndarray): the wave function
        """
        left_x = np.abs(psi[:, 1]) ** 2  # entire left boundary
        right_x = np.abs(psi[:, -2]) ** 2  # entire right boundary
        top_r = np.abs(psi[-2, :]) ** 2  # entire top boundary in radial direction
        tol = 1e-3
        if self.tof_nf:
            tol = 1e-5
        elif self.tof_ff:
            # TODO: Get with victor to see if value too wide; ticket: ALB-6344
            tol = 1e-7

        if any(i > tol for i in left_x):
            print(
                "Condensate hit left boundary at time:  ",
                self.times[-1] * self.qm_potential.sim_time_to_oqt,
            )
            self.hit_boundary = True
        if any(i > tol for i in right_x):
            print(
                "Condensate hit right boundary at time:  ",
                self.times[-1] * self.qm_potential.sim_time_to_oqt,
            )
            self.hit_boundary = True
        if any(i > tol for i in top_r):  # are any of the densities larger than tol?
            print(
                "Condensate hit top boundary at time:  ",
                self.times[-1] * self.qm_potential.sim_time_to_oqt,
            )
            self.hit_boundary = True

    @property
    def it_plot(self) -> dict:
        """
        Generate a simulation analog to an in-trap image from the Oqtant hardware.

        Returns:
            dict: data for generating an Image object(pixels, pixcal, rows, columns)
        """

        wavefunction = self.which_wavefunction_mode(self.times[-1])
        wavefunction.psi = self.psi_history[-1]  # we display the final result only

        # get the Y-integrated column denisty from wavefunction number / simulation area
        _, density_pixels, _ = wavefunction.column_densities

        density_pixels = (
            density_pixels.flatten()
            * wavefunction.three_d_grid.dx
            * wavefunction.three_d_grid.dz
        )

        # get conversion from simulation length to um
        pixcal = QMPotential.sim_length_to_oqt

        a_pixel_area = (
            pixcal**2 * wavefunction.three_d_grid.dz * wavefunction.three_d_grid.dx
        )  # * (pixcal) ** 2  # um2

        it_exp_pixcal = 0.344  # um

        # resonant cross section, D2 line, pi polarized light
        # REF: https://steck.us/alkalidata/rubidium87numbers.1.6.pdf
        sigma = 1.938e-1  # um2

        # convert density (atoms/length^2) to OD using the sim grid size and cross-section
        od_pixels = density_pixels * sigma * self.number_of_atoms / a_pixel_area

        # apply experimental imaging scaling to OD image
        # this is the average ratio between the number of atoms in experimental IT images vs TOF images
        # for heartbeat jobs due to inefficiencies in detection (i.e. magnetic field during imaging)
        # this necessarily destroys the atom number information in simulator generated OD images

        od_pixels *= self.EXPERIMENT_IT_TOF_OD_RATIO

        # half the length of the simulation grid axes, in um
        lx_half = (wavefunction.three_d_grid.Lx / 2) * pixcal
        ly_half = wavefunction.three_d_grid.Lr * pixcal

        # grids for building simulation interpolation function (um)
        sim_grid_col = np.linspace(-lx_half, lx_half, wavefunction.three_d_grid.Nx)
        sim_grid_row = np.linspace(-ly_half, ly_half, wavefunction.three_d_grid.Nr * 2)

        # build interpolation function for sim data
        f = interpolate.RegularGridInterpolator(
            (sim_grid_row, sim_grid_col),
            od_pixels.reshape(
                (wavefunction.three_d_grid.Nr * 2, wavefunction.three_d_grid.Nx)
            ),
        )

        # grids for sampling simulation in experiment resolution (um)
        exp_grid_col = np.arange(-lx_half, lx_half, it_exp_pixcal)
        exp_grid_row = np.arange(-ly_half, ly_half, it_exp_pixcal)

        exp_meshgrid_row, exp_meshgrid_col = np.meshgrid(exp_grid_row, exp_grid_col)

        points = np.array([exp_meshgrid_row.flatten(), exp_meshgrid_col.flatten()]).T

        # sample the interpolation function and reshape the output
        interpolated_sim = f(points).reshape(len(exp_grid_col), len(exp_grid_row))

        # how many pixels to pad the image?
        add_zeros_rows = math.ceil((148 - len(exp_grid_row)) / 2)
        add_zeros_cols = math.ceil((512 - len(exp_grid_col)) / 2)

        # pad the image to match IT_PLOT size
        padded_interp_sim = np.pad(
            interpolated_sim,
            [
                [add_zeros_cols, 512 - len(exp_grid_col) - add_zeros_cols],
                [add_zeros_rows, 148 - len(exp_grid_row) - add_zeros_rows],
            ],
        ).T

        # prepare values to be returned
        columns = len(padded_interp_sim[0])
        rows = len(padded_interp_sim)
        pixels_pad_interp_sim = list(padded_interp_sim.flatten())

        return {
            "it_plot": {
                "pixels": pixels_pad_interp_sim,
                "rows": rows,
                "columns": columns,
                "pixcal": it_exp_pixcal,
            }
        }

    @property
    def tof_output(self) -> dict:
        """
        Generate an simulation analog to a TOF image from the Oqtant hardware.

        Returns:
            dict: data for generating an Image object(pixels, pixcal, rows, columns)
        """

        # TODO refactor See; ALB-6345

        wavefunction = self.which_wavefunction_mode(self.times[-1])

        # generate and fit tof image with output class TF_dist_2D
        tof_exp_pixcal = 4.4  # um

        # get the Y-integrated column denisty from wavefunction number / simulation area
        _, density_pixels, _ = wavefunction.column_densities

        density_pixels = (
            density_pixels * wavefunction.three_d_grid.dx * wavefunction.three_d_grid.dz
        )

        # get conversion from simulation length to um
        pixcal = QMPotential.sim_length_to_oqt

        a_pixel_area = (
            pixcal**2 * wavefunction.three_d_grid.dz * wavefunction.three_d_grid.dx
        )  # um2

        # This sigma is sued for mF=2 -> mF=3; polarized light absorption cross-section
        # Ref: https://steck.us/alkalidata/rubidium87numbers.1.6.pdf
        sigma = 2.907e-1  # um2

        # convert density (atoms/length^2) to OD using the sim grid size and cross-section
        od_pixels = density_pixels * sigma * self.number_of_atoms / a_pixel_area

        int(len(od_pixels) / 2)
        int(len(od_pixels[0]) / 2)

        # half the length of the simulation grid axes, in um
        lx_half = (wavefunction.three_d_grid.Lx / 2) * pixcal
        lz_half = wavefunction.three_d_grid.Lr * pixcal

        # grids for building simulation interpolation function (um)
        sim_grid_col = np.linspace(-lx_half, lx_half, wavefunction.three_d_grid.Nx)
        sim_grid_row = np.linspace(-lz_half, lz_half, wavefunction.three_d_grid.Nr * 2)

        # build interpolation function for sim data
        f = interpolate.RegularGridInterpolator(
            (sim_grid_row, sim_grid_col),
            od_pixels,
            bounds_error=False,
            fill_value=0,
        )

        exp_grid_col = np.arange(
            -50 * tof_exp_pixcal, 50 * tof_exp_pixcal, tof_exp_pixcal
        )
        exp_grid_row = np.arange(
            -50 * tof_exp_pixcal, 50 * tof_exp_pixcal, tof_exp_pixcal
        )

        exp_meshgrid_row, exp_meshgrid_col = np.meshgrid(exp_grid_row, exp_grid_col)

        points = np.array([exp_meshgrid_row.flatten(), exp_meshgrid_col.flatten()]).T

        # sample the interpolation function and reshape the output
        interpolated_sim = f(points).reshape(len(exp_grid_col), len(exp_grid_row))

        # how many pixels to pad the image?
        add_zeros_rows = math.ceil((100 - len(exp_grid_row)) / 2)
        add_zeros_cols = math.ceil((100 - len(exp_grid_col)) / 2)

        # pad the image to match IT_PLOT size
        padded_interp_sim = np.pad(
            interpolated_sim,
            [
                [add_zeros_cols, 100 - len(exp_grid_col) - add_zeros_cols],
                [add_zeros_rows, 100 - len(exp_grid_row) - add_zeros_rows],
            ],
        ).T

        # prepare values to be returned
        columns = len(padded_interp_sim[0])
        rows = len(padded_interp_sim)
        pixels_interp_sim = list(padded_interp_sim.flatten())

        tof_image = Image(
            pixels=pixels_interp_sim,
            rows=rows,
            columns=columns,
            pixcal=tof_exp_pixcal,
        )

        x_points = padded_interp_sim[50]
        y_points = np.array(padded_interp_sim).T[50]

        d = []

        for i in range(100):
            d.append({"x": x_points[i], "y": y_points[i]})

        tof_x_slice = LineChart(points=d)
        tof_y_slice = LineChart(points=d)

        dummy_img = Image(
            pixels=[0 for i in range(100 * 100)], rows=2, columns=2, pixcal=1
        )

        dummy_fit = {
            "gaussian_od": 0,
            "gaussian_sigma_x": 0,
            "gaussian_sigma_y": 0,
            "tf_od": 0,
            "tf_x": 0,
            "tf_y": 0,
            "x_0": 0,
            "y_0": 0,
            "offset": 0,
        }

        output_dict = {
            "tof_image": tof_image,
            "tof_x_slice": tof_x_slice,
            "tof_y_slice": tof_y_slice,
            "mot_fluorescence_image": dummy_img,
            "tof_fit_image": dummy_img,
            "tof_fit": dummy_fit,
            "total_mot_atom_number": 0,
            "tof_atom_number": 0,
            "thermal_atom_number": 0,
            "condensed_atom_number": self.number_of_atoms,
            "temperature_nk": 0,
        }

        return output_dict

    def is_time_intrap(self, time: float) -> bool:
        """
        Checks if the time is when the condensate is still in trap.

        Args:
            time (float):   time in simulation units.
        Returns:
            Bool:  True means the system is in IT mode at that time.  False means it is not (it is in TOF mode).
        """
        test = time < self.qm_potential.lifetime
        if self.qm_potential.lifetime == 0.0:  # <-------------
            test = True  # edge case for lifetime = 0

        return test

    def is_time_near_field(self, time: float) -> bool:
        """
        Checks if the time is when the condensate is in near-field part of TOF.

        Args:
            time (float):   time in simulation (!) units.
        Returns:
            Bool:  True means the system is in NF of TOF mode at that time.
        """

        return (
            self.max_NF_time + self.qm_potential.lifetime
            > time
            >= self.qm_potential.lifetime
        )

    def is_time_far_field(self, time: float) -> bool:
        """
        Checks if the time is when the condensate is in far-field part of TOF.

        Args:
            time (float):   time in simulation (!) units.
        Returns:
            Bool:  True means the system is in FF of TOF mode at that time.
        """

        return time >= self.qm_potential.lifetime + self.max_NF_time

    def which_wavefunction_mode(self, time: float) -> WaveFunction:
        """
        Checks which mode the time corresponds to and returns the correct wavefunction class.

        Args:
            time (float):   time in simulation (!) units.
        Returns:
            WaveFunction:  instance of the wavefunction class in the correct mode
        """
        # checks which mode the simulator is in at that time.
        if self.is_time_intrap(time):
            wavefunction = self.wavefunction
            # update the potential to correct value in trap.
            self.qm_potential.update_potential(time)

        elif self.is_time_near_field(time):
            wavefunction = self.wavefunction_tof_nf

        elif self.is_time_far_field(time):
            wavefunction = self.wavefunction_tof_ff

        else:
            print("No wavefunction found for the requested timestep")
            raise ValueError("No wavefunction found for the requested timestep")

        return wavefunction

    def show_final_result(self) -> None:
        """
        Plot the density at the end of the simulation in cylindrical coordinates.
        Useful coordinates for diagonising issues but not to be returned to the user.
        """

        wavefunction = self.which_wavefunction_mode(self.times[-1])
        wavefunction.psi = self.psi_history[-1]  # we display the final result only

        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
        density = (
            wavefunction.density / QMPotential.sim_length_to_oqt**3
        )  # atoms / micron^3
        ax.plot_surface(
            self.qm_potential.sim_length_to_oqt * wavefunction.three_d_grid.r,
            self.qm_potential.sim_length_to_oqt * wavefunction.three_d_grid.x,
            density,
            cmap="viridis",
        )
        plt.title("Density (Atoms/$\\mu$m$^3$) at final time")
        ax.set_xlabel("$R$-Position ($\\mu$m)")
        ax.set_ylabel("$X$-Position ($\\mu$m)")

    def convert_timesteps(self, timesteps: list) -> np.ndarray:
        """
        Convert a list of arbitrary times (in oqtant units) to a list of simulator timestep indices.
        Simulation must already have been evaluated.

        Args:
            timesteps (list): list of times in oqtant units

        Returns:
            ndarray : array of simulator timestep indexes (not the values of the timesteps)

        """

        if self.times is None:
            raise SimValueError("times not set")

        np.ceil(QMPotential.sim_time_to_oqt * self.qm_potential.lifetime)
        np.ceil(
            QMPotential.sim_time_to_oqt
            * (self.qm_potential.lifetime + self.max_NF_time)
        )

        time_end_oqt = np.ceil(
            QMPotential.sim_time_to_oqt
            * (self.qm_potential.time_of_flight + self.qm_potential.lifetime)
        )
        last_timestep_idx = self.times[-1]

        np_timesteps = np.array(timesteps, dtype=float)
        # convert input time in msec to simulation units
        np_timesteps *= QMPotential.oqt_time_to_sim

        # if the user has requested the end time of the simulation exactly, adjust this to the last saved timestep

        for i in range(len(timesteps)):
            oqt_time = timesteps[i]
            if oqt_time == time_end_oqt:
                np_timesteps[i] = last_timestep_idx

        if (
            np.max(np_timesteps) > self.times[-1]
            or np.min(np_timesteps) < self.times[0]
        ):
            raise SimValueError("timesteps outside of job window")

        if np.min(np_timesteps) < 0:
            raise SimValueError("timesteps must be positive")

        return np_timesteps

    def get_grids(self, time_ms: float) -> tuple:
        """
        Returns the x-grid in microns at a user specified time.

        Args:
            time_ms (float):  time (in milliseconds)
        Returns:
            tuple (dx, Lx, x_1d) : the simulation resolution, length in x-direction, and length array in x-direction
        """
        [sim_time] = self.convert_timesteps([time_ms])

        i = np.searchsorted(self.times, sim_time, side="left", sorter=None)

        wavefunction = self.which_wavefunction_mode(self.times[i])
        wavefunction.psi = self.psi_history[i]  # load a snapshot of psi

        oqt_length = self.qm_potential.sim_length_to_oqt
        Lx = wavefunction.three_d_grid.Lx * oqt_length
        x_1d = wavefunction.three_d_grid.x_1d * oqt_length
        dx = wavefunction.three_d_grid.dx * oqt_length

        return dx, Lx, x_1d

    def get_times(self):
        """
        Returns the an array of  times in oqtant units

        Returns:
            array times : the timesteps in oqtant units (milliseconds)
        """
        times = np.array(self.times) * self.qm_potential.sim_time_to_oqt

        return times

    def get_column_densities(self, time_ms: float) -> tuple:
        """
        Returns the column densities and slices of the condensate in cartesian coordinates
        for an arbitrary time.
        In correct coordinates to be returned to the user.

        Args:
            time_ms (float):  time (in milliseconds)
        Returns:
            tuple (column_zy, column_zx, slice_y, slice_x) : the column densitites and slices at the desired time
        """
        [sim_time] = self.convert_timesteps([time_ms])

        i = np.searchsorted(self.times, sim_time, side="left", sorter=None)

        wavefunction = self.which_wavefunction_mode(self.times[i])
        wavefunction.psi = self.psi_history[i]  # load a snapshot of psi

        # load the column densities
        (
            column_zy,
            column_zx,
            profiles,
        ) = wavefunction.column_densities

        column_zy *= self.number_of_atoms
        column_zx *= self.number_of_atoms

        slice_x = column_zx[int(wavefunction.three_d_grid.Nr - 1), :]
        slice_y = column_zy[int(wavefunction.three_d_grid.Nr - 1), :]

        return column_zy, column_zx, slice_y, slice_x

    def show_density_cylindrical(self, times_ms: list, figsize=(15, 7)) -> None:
        """
        Plots the density profile of the condensate in cylindrical coordinates
        for an input array of times.
        Useful coordinates for diagonising issues but not to be returned to the user.

        Args:
            times_ms (list):  a list of times (in milliseconds/oqtant units)
            figsize (tuple): size of output figure. default = (15,7)
        Returns:
            None
        """
        # convert into simulation times
        converted_timesteps = self.convert_timesteps(times_ms)

        # convert to simulation indices for the nearest timestep, loop over timesteps
        for time in converted_timesteps:
            i = np.searchsorted(self.times, time, side="left", sorter=None)

            wavefunction = self.which_wavefunction_mode(self.times[i])
            fig = plt.figure(figsize=figsize)

            # Add first subplot in a plot with two rows and 2 columns.
            ax1 = fig.add_subplot(111)
            wavefunction.psi = self.psi_history[i]
            density = (
                wavefunction.density
                * self.number_of_atoms
                / QMPotential.sim_length_to_oqt**3
            )  # atoms / micron^3
            img1 = ax1.imshow(
                np.concatenate(
                    (np.flip(density, axis=0), density)
                ),  # mirror the condensate over the z-axis
                extent=[
                    wavefunction.three_d_grid.x[0, 0]
                    * self.qm_potential.sim_length_to_oqt,
                    wavefunction.three_d_grid.x[0, -1]
                    * self.qm_potential.sim_length_to_oqt,
                    -wavefunction.three_d_grid.r[-1, 0]
                    * self.qm_potential.sim_length_to_oqt,
                    wavefunction.three_d_grid.r[-1, 0]
                    * self.qm_potential.sim_length_to_oqt,
                ],
                aspect="auto",
                cmap="viridis",
            )
            oqt_time = self.times[i] * self.qm_potential.sim_time_to_oqt
            plt.title("Density (Atoms/$\\mu$m$^3$) at t = %.2f ms" % oqt_time)
            ax1.set_xlabel("$X$-Position ($\\mu$m)")
            ax1.set_ylabel("$R$-Position ($\\mu$m)")
            fig.colorbar(img1, ax=ax1, orientation="vertical")

    def show_column_densities(
        self, times_ms: list, slices: bool = True, figsize=(15, 7)
    ):
        """
        Plots the column densities and slices of the condensate in cartesian coordinates
        for an input array of times.
        In correct coordinates to be returned to the user.

        Args:
            times_ms (list):  a list of times (in milliseconds)
            slices (bool): plot or not to plot slices
            figsize (tuple): size of output figure. default = (15,7)
        Returns:
            None
        """
        converted_timesteps = self.convert_timesteps(times_ms)

        for time in converted_timesteps:

            i = np.searchsorted(self.times, time, side="left", sorter=None)
            wavefunction = self.which_wavefunction_mode(self.times[i])

            wavefunction.psi = self.psi_history[i]  # load a snapshot of psi

            # load the column densities
            (
                column_zy,
                column_zx,
                profiles,
            ) = wavefunction.column_densities
            slice_x_2 = column_zx[int(wavefunction.three_d_grid.Nr - 1), :]
            slice_z = column_zy[int(wavefunction.three_d_grid.Nr - 1), :]
            slice_z_zy, slice_z_zx, slice_x = profiles

            fig = plt.figure(figsize=figsize)

            # column densities
            # Add first subplot in a plot with two rows and 2 columns.
            ax1 = fig.add_subplot(221)
            im1 = ax1.imshow(
                self.number_of_atoms
                * column_zy
                / QMPotential.sim_length_to_oqt**2,  # atoms / micron^2,
                extent=[
                    -wavefunction.three_d_grid.Lr * self.qm_potential.sim_length_to_oqt,
                    wavefunction.three_d_grid.Lr * self.qm_potential.sim_length_to_oqt,
                    -wavefunction.three_d_grid.Lr * self.qm_potential.sim_length_to_oqt,
                    wavefunction.three_d_grid.Lr * self.qm_potential.sim_length_to_oqt,
                ],
                aspect="auto",
                cmap="viridis",
            )
            oqt_time = self.times[i] * self.qm_potential.sim_time_to_oqt
            plt.title("Column densities at t = %.2f ms" % oqt_time)
            plt.xlabel("$Y$-Position ($\\mu$m)")
            plt.ylabel("$Z$-Position ($\\mu$m)")
            cbar1 = fig.colorbar(im1, ax=ax1, orientation="vertical")
            cbar1.ax.set_ylabel("Density (Atoms/$\\mu$m$^2$)")

            ax2 = fig.add_subplot(222)
            im2 = ax2.imshow(
                self.number_of_atoms
                * column_zx
                / QMPotential.sim_length_to_oqt**2,  # atoms / micron^2,
                extent=[
                    -self.qm_potential.sim_length_to_oqt
                    * wavefunction.three_d_grid.Lx
                    / 2,
                    self.qm_potential.sim_length_to_oqt
                    * wavefunction.three_d_grid.Lx
                    / 2,
                    -self.qm_potential.sim_length_to_oqt * wavefunction.three_d_grid.Lr,
                    self.qm_potential.sim_length_to_oqt * wavefunction.three_d_grid.Lr,
                ],
                aspect="auto",
                cmap="viridis",
            )
            plt.xlabel("$X$-Position ($\\mu$m)")
            plt.ylabel("$Z$-Position ($\\mu$m)")
            cbar2 = fig.colorbar(im2, ax=ax2, orientation="vertical")
            cbar2.ax.set_ylabel("Density (Atoms/$\\mu$m$^2$)")

            # slices
            if slices:
                ax3 = fig.add_subplot(223)
                pos1 = ax1.get_position()
                pos3 = ax3.get_position()

                ax3.plot(
                    wavefunction.three_d_grid.z[0, :]
                    * self.qm_potential.sim_length_to_oqt,
                    self.number_of_atoms
                    * slice_z
                    / QMPotential.sim_length_to_oqt,  # atoms / micron
                )
                plt.title("Slices at t = %.2f ms" % oqt_time)
                ax3.set_xlabel("$Z$-Position ($\\mu$m)")
                ax3.set_ylabel("Density (Atoms/$\\mu$m)")
                ax3.set_position([pos1.x0, pos3.y0 * 0.5, pos1.width, pos3.height])

                #
                ax4 = fig.add_subplot(224)
                pos2 = ax2.get_position()
                pos4 = ax4.get_position()

                ax4.plot(
                    wavefunction.three_d_grid.x[0, :]
                    * self.qm_potential.sim_length_to_oqt,
                    self.number_of_atoms
                    * slice_x_2
                    / QMPotential.sim_length_to_oqt,  # atoms / micron,
                )
                ax4.set_xlabel("$X$-Position ($\\mu$m)")
                ax4.set_ylabel("Density (Atoms/$\\mu$m)")
                ax4.set_position([pos2.x0, pos4.y0 * 0.5, pos2.width, pos4.height])

            plt.show()

    def show_phase(self, times_ms: list, figsize=(10, 7)) -> None:
        """
        Plot the phase for a given list of timesteps
        This can only be displayed in cylindrical coordinates.
        It is a helpful tool still for the user.  The aspect ratio is still a bit weird.

        Args:
            times_ms (list): List of times to display
            figsize (tuple): size of output figure. default = (10,7)
        Returns:
            None
        """
        converted_timesteps = self.convert_timesteps(times_ms)

        for time in converted_timesteps:
            # convert to simulation indices for the nearest timestep, loop over timesteps
            i = np.searchsorted(self.times, time, side="left", sorter=None)

            wavefunction = self.which_wavefunction_mode(self.times[i])

            wavefunction.psi = self.psi_history[i]

            plt.figure(figsize=figsize)
            plt.imshow(
                np.concatenate(
                    (np.flip(wavefunction.phase, axis=0), wavefunction.phase)
                ),  # needs to be flipped to plot x on x axis and r on y axis.
                extent=[
                    wavefunction.three_d_grid.x[0, 0]
                    * self.qm_potential.sim_length_to_oqt,
                    wavefunction.three_d_grid.x[0, -1]
                    * self.qm_potential.sim_length_to_oqt,
                    -wavefunction.three_d_grid.r[-1, 0]
                    * self.qm_potential.sim_length_to_oqt,
                    wavefunction.three_d_grid.r[-1, 0]
                    * self.qm_potential.sim_length_to_oqt,
                ],
                aspect="auto",
                cmap="RdYlBu",
            )
            plt.colorbar(label="Phase (Radians)")
            plt.clim(-np.pi, np.pi)
            oqt_time = self.times[i] * self.qm_potential.sim_time_to_oqt
            plt.title("Phase at t = %.2f ms" % oqt_time)
            plt.xlabel("$X$-Position ($\\mu$m)")
            plt.ylabel("$R$-Position ($\\mu$m)")
            plt.show()

    def show_current(self, times_ms: list, figsize=(10, 7)) -> None:
        """
        Plot the flow for a given list of timesteps
        two separate subplots

        Args:
            times_ms (list): List of times (ms) at which to display current
            figsize (tuple): size of output figure. default = (10,7)

        Returns:
            None
        """
        converted_timesteps = self.convert_timesteps(times_ms)

        for time in converted_timesteps:
            i = np.searchsorted(self.times, time, side="left", sorter=None)

            wavefunction = self.which_wavefunction_mode(self.times[i])
            wavefunction.psi = self.psi_history[i]

            plt.figure(1, figsize=figsize)
            plt.plot(
                wavefunction.three_d_grid.x_1d * self.qm_potential.sim_length_to_oqt,
                self.number_of_atoms
                * wavefunction.current
                / self.qm_potential.sim_time_to_oqt,
            )
            oqt_time = self.times[i] * self.qm_potential.sim_time_to_oqt
            plt.title("Current at t = %.2f ms" % oqt_time)
            plt.xlabel("$X$-Position ($\\mu$m)")
            plt.ylabel("Atom Current (# of Atoms / msec)")
            plt.show()

    def animate_current(
        self, frame_interval: int = 1, y_limit=number_of_atoms
    ) -> FuncAnimation:
        """
        Animates the density profiles and change in potentials over time.  This is an integrated profile along
        the x-direction and is different from a single slice.
        Args:
            frame_interval (int): number of frames to skip each interval, determines smoothness. default =1
            current_bound (float):  adjustable bound on the y-axis in the case that the current exceeds the default.
        Returns:
            FuncAnimation: an animation of the profile along the x-direction.
        """
        fig, ax1 = plt.subplots(1, 1)

        self.wavefunction.psi = self.psi_history[0]

        (line1,) = ax1.plot([], [], lw=2, color="b", label="Atom Current - X")
        line = [line1]

        ax1.set_ylim(-y_limit, y_limit)
        ax1.set_xlim(
            -QMPotential.sim_length_to_oqt * self.wavefunction.three_d_grid.Lx / 2,
            QMPotential.sim_length_to_oqt * self.wavefunction.three_d_grid.Lx / 2,
        )

        ax1.grid()
        ax1.legend(loc="upper right")
        ax1.set_xlabel("$X$ ($\\mu$m)")
        ax1.set_ylabel("Atom Current (# of Atoms / msec)")
        fig.suptitle("t = 0 ms", x=0.5, y=0.85)

        fig.tight_layout()
        plt.close()

        def init_animation():
            """Callable for matplotlib used for initializing animation"""

            line[0].set_data([], [])
            return line

        def update(frame):
            """Callable for matplotlib used for updating the wf through the animation."""
            self.qm_potential.update_potential(self.times[frame])
            self.wavefunction.psi = self.psi_history[frame]

            current = (
                self.number_of_atoms
                * self.wavefunction.current
                / self.qm_potential.sim_time_to_oqt
            )

            x1 = (
                self.wavefunction.three_d_grid.x[0]
                * self.qm_potential.sim_length_to_oqt
            )
            y1 = current

            oqt_time = self.times[frame] * self.qm_potential.sim_time_to_oqt
            fig.suptitle("t = %.2f ms" % oqt_time, x=0.5, y=0.85)

            line[0].set_data(x1, y1)
            return (line,)

        anim = FuncAnimation(
            fig,
            update,
            init_func=init_animation,
            frames=range(0, self.last_it_time_idx, frame_interval),
            interval=50 + 100 * frame_interval,
            blit=False,
        )

        return anim

    def animate_phase(
        self, frame_interval: int = 1, show_potential: bool = True, figsize=(8, 3)
    ):
        """
        Animate the change in phase

        Args:
            frame_interval (int): number of frames to skip each interval, determines smoothness. default =1
            show_potential (bool): whether or not to show the potential on the animation. default = True
            figsize (tuple): size of the output figure. default = (8,3)

        Returns:
            FuncAnimation: an animation of the profile along the x-direction.
        """
        fig, ax = plt.subplots(figsize=figsize)

        self.qm_potential.update_potential(self.times[0])
        self.wavefunction.psi = self.psi_history[0]

        phase = self.wavefunction.phase

        extent = [
            self.wavefunction.three_d_grid.x[0, 0]
            * self.qm_potential.sim_length_to_oqt,
            self.wavefunction.three_d_grid.x[0, -1]
            * self.qm_potential.sim_length_to_oqt,
            -self.wavefunction.three_d_grid.r[-1, 0]
            * self.qm_potential.sim_length_to_oqt,
            self.wavefunction.three_d_grid.r[-1, 0]
            * self.qm_potential.sim_length_to_oqt,
        ]
        im = ax.imshow(
            np.concatenate((np.flip(phase, axis=0), phase)),
            extent=extent,
            vmin=-np.pi,
            vmax=np.pi,
            cmap="RdYlBu",
        )

        if show_potential:
            cont1 = ax.contour(
                self.wavefunction.three_d_grid.x_1d
                * self.qm_potential.sim_length_to_oqt,
                np.concatenate(
                    (
                        -np.flip(self.wavefunction.three_d_grid.r_1d, axis=0),
                        self.wavefunction.three_d_grid.r_1d,
                    )
                )
                * self.qm_potential.sim_length_to_oqt,
                self.qm_potential.potential_to_cartesian_oqt_units(),
                colors="white",
            )

        ax.set_xlabel("X (um)")
        ax.set_ylabel("R (um)")
        fig.suptitle("t = 0 ms", x=0.5, y=0.85)
        plt.colorbar(im, ax=ax, label="Phase (Radians)")
        plt.close()

        def update(frame):
            if show_potential:
                nonlocal cont1
                for c in cont1.collections:
                    c.remove()
            self.qm_potential.update_potential(self.times[frame])
            self.wavefunction.psi = self.psi_history[frame]
            phase = self.wavefunction.phase
            im.set_array(np.concatenate((np.flip(phase, axis=0), phase)))
            oqt_time = self.times[frame] * self.qm_potential.sim_time_to_oqt
            fig.suptitle("t = %.2f ms" % oqt_time, x=0.5, y=0.85)
            cont1 = ax.contour(
                self.wavefunction.three_d_grid.x_1d
                * self.qm_potential.sim_length_to_oqt,
                np.concatenate(
                    (
                        -np.flip(self.wavefunction.three_d_grid.r_1d, axis=0),
                        self.wavefunction.three_d_grid.r_1d,
                    )
                )
                * self.qm_potential.sim_length_to_oqt,
                self.qm_potential.potential_to_cartesian_oqt_units(),
                colors="white",
            )

            if show_potential:
                return im, cont1
            return im

        anim = FuncAnimation(
            fig,
            update,
            frames=range(0, self.last_it_time_idx, frame_interval),
            interval=50 + 100 * frame_interval,
            blit=False,
        )

        return anim

    def animate_density(self, frame_interval=1, figsize=(8, 3), show_potential=True):
        """
        Animates the change in density and potential over time

        Args:
            frame_interval (int): number of frames to skip each interval, determines smoothness. default = 1
            show_potential (bool): whether or not to show the potential on the animation. default = True
            figsize (tuple): size of the output figure. default = (8,3)

        Returns:
            FuncAnimation: an animation of the profile along the x-direction.

        """
        fig, ax = plt.subplots(figsize=(figsize))

        wavefunction = self.wavefunction

        _, density, _ = wavefunction.column_densities

        extent = [
            wavefunction.three_d_grid.x[0, 0] * self.qm_potential.sim_length_to_oqt,
            wavefunction.three_d_grid.x[0, -1] * self.qm_potential.sim_length_to_oqt,
            -wavefunction.three_d_grid.y[-1, 0] * self.qm_potential.sim_length_to_oqt,
            wavefunction.three_d_grid.y[-1, 0] * self.qm_potential.sim_length_to_oqt,
        ]

        # vmin and vmax normalize the colorbar, making the cloud more visible
        im = ax.imshow(
            self.number_of_atoms
            * density
            / QMPotential.sim_length_to_oqt**2,  # atoms / micron^2,
            extent=extent,
            cmap="viridis",
        )  # normalize in here with vmin and vmax
        if show_potential:
            cont1 = ax.contour(
                wavefunction.three_d_grid.x_1d * self.qm_potential.sim_length_to_oqt,
                np.concatenate(
                    (
                        -np.flip(wavefunction.three_d_grid.r_1d, axis=0),
                        wavefunction.three_d_grid.r_1d,
                    )
                )
                * self.qm_potential.sim_length_to_oqt,
                self.qm_potential.potential_to_cartesian_oqt_units(),
                colors="white",
            )

        ax.set_xlabel("X (um)")
        ax.set_ylabel("Y (um)")
        fig.suptitle("t = 0 ms", x=0.5, y=0.85)

        plt.colorbar(im, ax=ax, label="Density (Atoms/$\\mu$m$^2$)")
        plt.close()

        def update(frame):
            self.qm_potential.update_potential(self.times[frame])
            wavefunction.psi = self.psi_history[frame]
            _, density, _ = wavefunction.column_densities
            im.set_array(
                self.number_of_atoms * density / QMPotential.sim_length_to_oqt**2
            )  # atoms / micron^2)
            oqt_time = self.times[frame] * self.qm_potential.sim_time_to_oqt
            fig.suptitle("t = %.2f ms" % oqt_time, x=0.5, y=0.85)
            if show_potential:
                nonlocal cont1
                for c in cont1.collections:
                    c.remove()
                cont1 = ax.contour(
                    wavefunction.three_d_grid.x_1d
                    * self.qm_potential.sim_length_to_oqt,
                    np.concatenate(
                        (
                            -np.flip(wavefunction.three_d_grid.r_1d, axis=0),
                            wavefunction.three_d_grid.r_1d,
                        )
                    )
                    * self.qm_potential.sim_length_to_oqt,
                    self.qm_potential.potential_to_cartesian_oqt_units(),
                    colors="white",
                )

            if show_potential:
                return im, cont1  # , cont2, cont3, cont4
            else:
                return im

        anim = FuncAnimation(
            fig,
            update,
            frames=range(0, self.last_it_time_idx, frame_interval),
            interval=50 + 100 * frame_interval,
            blit=False,
        )

        return anim

    def animate_profiles(
        self,
        frame_interval=1,
        y_limit=number_of_atoms / (QMPotential.sim_length_to_oqt * 5),
    ):
        """
        Animates the density profiles and change in potentials over time.  This is an integrated profile along
        the x-direction and is different from a single slice(!).

        Args:
            frame_interval (int): number of frames to skip each interval, determines smoothness. default = 1
            y_axis (float):  optional arguement to adjust the y-axis in the event that the density is to large.

        Returns:
            FuncAnimation: an animation of the profile along the x-direction.

        """
        fig, ax1 = plt.subplots(1, 1)

        self.wavefunction.psi = self.psi_history[0]

        profile_x, _ = self.wavefunction.density_profiles

        (line1,) = ax1.plot([], [], lw=2, color="b", label="Density Profile - X")
        line = [line1]

        ax1.set_ylim(0, y_limit)
        ax1.set_xlim(
            -QMPotential.sim_length_to_oqt * self.wavefunction.three_d_grid.Lx / 2,
            QMPotential.sim_length_to_oqt * self.wavefunction.three_d_grid.Lx / 2,
        )

        ax1.grid()
        ax1.legend(loc="upper right")
        ax1.set_xlabel("X (um)")
        ax1.set_ylabel("Density (Atoms/$\\mu$m$^2$)")
        fig.suptitle("t = 0 ms", x=0.5, y=0.85)

        fig.tight_layout()
        plt.close()

        def init_animation():
            """Callable for matplotlib used for initializing animation"""
            line[0].set_data([], [])
            return line

        def update(frame):
            """Callable for matplotlib used for updating the wf through the animation."""
            self.qm_potential.update_potential(self.times[frame])
            self.wavefunction.psi = self.psi_history[frame]

            profile_x, _ = self.wavefunction.density_profiles

            oqt_time = self.times[frame] * self.qm_potential.sim_time_to_oqt
            fig.suptitle("t = %.2f ms" % oqt_time, x=0.5, y=0.85)

            x1 = (
                self.wavefunction.three_d_grid.x[0]
                * self.qm_potential.sim_length_to_oqt
            )

            y1 = self.number_of_atoms * profile_x / QMPotential.sim_length_to_oqt

            line[0].set_data(x1, y1)
            return (line,)

        anim = FuncAnimation(
            fig,
            update,
            init_func=init_animation,
            frames=range(0, self.last_it_time_idx, frame_interval),
            interval=50 + 100 * frame_interval,
            blit=False,
        )

        return anim



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from bert_schemas import job as job_schema

from oqtant.simulator.three_dim import ThreeDimGrid
from bert_schemas.timing import decimal_to_float


class QMPotential:
    """
    'QMPotential' represents the quantum matter object potential (combination of magnetic trap/snapshot/barriers)
    in simulation units. Contains a 2D array of the potential energy in the simulation at a given time.
    """

    # natural units of Oqtant
    weak_trap_freq = 2 * np.pi * 50  # weak axis trap frequency
    m = 87 * 1.66054e-27  # rb87 mass
    h = 6.626e-34  # planck's constant
    msec = 1e-3
    microns = 1e-6
    khz = 1e3
    hbar = h / 2 / np.pi  # hbar

    # natural units of the simulation (in SI) #CHECKED
    sim_time_unit = 1 / weak_trap_freq  # 3.18 milliseconds = 1 simulation time
    sim_length_unit = np.sqrt(
        hbar / m / weak_trap_freq
    )  # 1.52 microns = 1 simulation length
    sim_energy_unit = hbar * weak_trap_freq

    # convert simulation units to Oqtant units # Checked
    sim_length_to_oqt = (
        sim_length_unit / microns
    )  # convert a simulation length to microns.  1.524 microns
    sim_time_to_oqt = (
        sim_time_unit / msec
    )  # convert a simulation time to microseconds.  3.183 mseconds
    oqt_time_to_sim = 1 / sim_time_to_oqt

    w0x = 1.0  # 50hz; trap angular frequency; set natural units
    w0r = 8.0  # 400hz; trap angular frequency

    def __init__(self, quantum_matter):
        self.quantum_matter = quantum_matter
        self.three_d_grid = ThreeDimGrid()
        self.potential = None

    # convert units of quantum matter object to simulation relevant quantities
    @property
    def lifetime(self) -> float:
        """Returns the lifetime of the simulation in simulation time units."""
        return (
            decimal_to_float(self.quantum_matter.lifetime) * QMPotential.msec / QMPotential.sim_time_unit
        )

    @property
    def time_of_flight(self) -> float:
        """If image type is in trap returns zero. Else calculate and return time of flight in the units of the sim"""
        if self.quantum_matter.image == job_schema.ImageType.IN_TRAP:
            return 0.0

        return (
            decimal_to_float(self.quantum_matter.time_of_flight)
            * QMPotential.msec
            / QMPotential.sim_time_unit
        )

    def update_potential(self, time: np.ndarray, clip: bool = False) -> None:
        """
        Function to query the potential at a specific simulation time (sim units), from the Oqtant
        quantum matter object. Potential is updated with the magnetic trap

        Updates property self.potential

        Args:
            time (np.ndarray[float]):  time (in simulation units)
            clip (boolean): whether to clip

        Returns:
            None
        """
        if time > self.lifetime + self.time_of_flight:
            raise ValueError(
                "Potential update requested outside time bounds of the simulation"
            )

        # time: sim -> msec # Checked
        time *= QMPotential.sim_time_unit / QMPotential.msec

        # potential in x-direction (in kHz)
        potential_1d_x = np.array(
            self.quantum_matter.get_potential(
                time,
                self.three_d_grid.x_1d
                * QMPotential.sim_length_unit
                / QMPotential.microns,
            )
        )

        # potential in z-direction (in sim. units) # Checked
        potential_1d_x *= QMPotential.h * QMPotential.khz / QMPotential.sim_energy_unit

        # potential in r-direction (in sim. units) # Checked
        potential_1d_r = 0.5 * QMPotential.w0r**2 * self.three_d_grid.r_1d**2

        # meshgrid in r and x directions (simulation units)
        pot_x, pot_r = np.meshgrid(potential_1d_x, potential_1d_r)

        # total potential: simulation units
        if clip:
            self.potential = np.clip(
                pot_x + pot_r, 0.0, 400.0
            )  # clip energies above 20 khz.  can be raised...

        else:
            self.potential = pot_x + pot_r

    def potential_to_cartesian_oqt_units(self) -> np.ndarray:
        """
        Convert the potential object self.potential to cartesian coordinates and oqtant units (microns, kHz).

        Returns:
            array: the converted potential in cartesian coordinates and oqtant units
        """

        # convert potential to oqtant units from sim energy units
        potential_khz = (
            self.potential
            * QMPotential.sim_energy_unit
            / (QMPotential.h * QMPotential.khz)
        )

        return np.concatenate((np.flip(potential_khz, axis=0), potential_khz))



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Any
from uuid import UUID

from bert_schemas import job as job_schema
from dateutil import parser as date_parser
from dateutil import tz

SIG_ABS = 0.297


def print_keys(subject: Any, indent: int = 0, drill_lists: bool = False) -> None:
    """Print the keys of a nested dictionary or list

    Args:
        subject (Any): The subject to print the keys of
        indent (int, optional): The number of spaces to indent. Defaults to 0
        drill_lists (bool, optional): Whether to drill into lists. Defaults to False
    """
    if isinstance(subject, dict):
        for key, value in subject.items():
            print(f"{' ' * indent}- {key}")
            if isinstance(value, dict):
                print_keys(value, indent=indent + 2, drill_lists=drill_lists)
            if isinstance(value, list) and drill_lists:
                for list_value in value:
                    if isinstance(list_value, dict):
                        print_keys(
                            list_value, indent=indent + 2, drill_lists=drill_lists
                        )


class OqtantJob(job_schema.JobBase):
    """A class that represents a job submitted to Oqtant"""

    external_id: UUID | None = None
    time_submit: str | datetime | None = None
    time_start: str | datetime | None = None
    time_complete: str | datetime | None = None

    @property
    def truncated_name(self) -> str:
        """Property to truncate a job's name

        Returns:
            str: The truncated job name
        """
        if len(self.name) > 40:
            return f"{self.name[:37]}..."
        return self.name

    @property
    def id(self) -> UUID:
        """Property to get the id of an OqtantJob object

        Returns:
            uuid.UUID: The id of the OqtantJob object
        """
        return self.external_id

    @property
    def formatted_time_submit(self) -> str:
        """Property to format the job submit datetime and ensure it is in caller's local timezone

        Returns:
            str: The formatted datetime string
        """
        return self.format_datetime(self.time_submit)

    @staticmethod
    def format_datetime(datetime_value: str | datetime) -> str:
        """Method to format any datetime and ensure it is in caller's local timezone

        Args:
            datetime_value (str | datetime.datetime): The datetime value to format

        Returns:
            str: The formatted datetime string
        """
        try:
            parsed_datetime = date_parser.parse(datetime_value)
            parsed_datetime = parsed_datetime.replace(tzinfo=tz.tzutc())
            parsed_datetime = parsed_datetime.astimezone(tz.tzlocal())
        except Exception:
            parsed_datetime = datetime_value
        return parsed_datetime.strftime("%d %b %Y, %H:%M:%S")

    @property
    def input_fields(self) -> None:
        """Property to print out all of the input fields for an OqtantJob"""
        print_keys(self.input.model_dump(), drill_lists=True)

    @property
    def input(self) -> job_schema.InputValues:
        """Property to get the input values for the current run of an OqtantJob

        Returns:
            bert_schemas.job.InputValues: The input values for the current run
        """
        return self.inputs[0].values

    @property
    def lifetime(self) -> int:
        """Property to get the lifetime value for the current run

        Returns:
            int: The lifetime value for the current run
        """
        return self.input.end_time_ms

    @property
    def run(self) -> int:
        """Property to get the current run value for the job's input

        Returns:
            int: The current run value for the job's input
        """
        return self.inputs[0].run

    def add_notes_to_input(self, notes: str) -> None:
        """Method to add notes to the current run

        Args:
            notes (str): The notes to add to the input
        """
        self.inputs[0].notes = notes



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

ultracold_matter_job = {
    "name": "Example Ultracold Matter Generator Job",
    "job_type": "BEC",
    "inputs": [
        {
            "values": {
                "time_of_flight_ms": 8.0,
                "image_type": "TIME_OF_FLIGHT",
                "end_time_ms": 0.0,
                "rf_evaporation": {
                    "frequencies_mhz": [17.0, 8.0, 4.0, 1.2, 0.045],
                    "powers_mw": [500.0, 500.0, 475.0, 360.0, 220.0],
                    "interpolation": "LINEAR",
                    "times_ms": [-1600, -1200, -800, -400, 0],
                },
                "optical_barriers": None,
                "optical_landscape": None,
                "lasers": None,
            },
        }
    ],
}

barrier_manipulator_job = {
    "name": "Example Barrier Manipulator Job",
    "job_type": "BARRIER",
    "inputs": [
        {
            "values": {
                "time_of_flight_ms": 7.0,
                "image_type": "IN_TRAP",
                "end_time_ms": 20.0,
                "rf_evaporation": {
                    "powers_mw": [500.0, 500.0, 475.0, 360.0, 220.0],
                    "frequencies_mhz": [17.0, 8.0, 4.0, 1.2, 0.1],
                    "interpolation": "LINEAR",
                    "times_ms": [-1600, -1200, -800, -400, 0],
                },
                "optical_barriers": [
                    {
                        "heights_khz": [
                            0.0,
                            5.0,
                            10.0,
                            15.0,
                            20.0,
                            25.0,
                            25.0,
                            25.0,
                            25.0,
                            25.0,
                            25.0,
                        ],
                        "positions_um": [
                            -10.0,
                            -10.0,
                            -10.0,
                            -10.0,
                            -10.0,
                            -10.0,
                            -10.0,
                            -10.0,
                            -10.0,
                            -10.0,
                            -10.0,
                        ],
                        "interpolation": "OFF",
                        "widths_um": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                        "times_ms": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                        "shape": "GAUSSIAN",
                    },
                    {
                        "heights_khz": [
                            0.0,
                            5.0,
                            10.0,
                            15.0,
                            20.0,
                            25.0,
                            30.0,
                            35.0,
                            40.0,
                            40.0,
                            40.0,
                        ],
                        "positions_um": [
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                            10.0,
                        ],
                        "interpolation": "OFF",
                        "widths_um": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                        "times_ms": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                        "shape": "GAUSSIAN",
                    },
                ],
                "optical_landscape": None,
                "lasers": None,
            }
        },
    ],
}



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class JobError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class JobReadError(JobError):
    pass


class JobWriteError(JobError):
    pass


class OqtantError(Exception):
    def __init__(self, message=None):
        self.message = message
        super().__init__(self.message)


class OqtantAuthorizationError(OqtantError):
    pass


class OqtantTokenError(OqtantError):
    pass


class OqtantRequestError(OqtantError):
    pass


class OqtantJobError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class OqtantJobUnsupportedTypeError(OqtantJobError):
    pass


class OqtantJobValidationError(OqtantJobError):
    pass


class OqtantJobParameterError(OqtantJobError):
    pass


class OqtantJobListLimitError(OqtantJobError):
    pass


class SimValueError(ValueError):
    pass


class SimSubmitError(Exception):
    def __init__(self):
        super().__init__("Cannot submit simulations as a batch.")


class JobPlotFitError(OqtantJobError):
    def __init__(self):
        super().__init__("PLOT FIT RESULTS: failed to generate model from provided parameters")


class JobPlotFitMismatchError(OqtantJobError):
    def __init__(self):
        super().__init__("PLOT FIT RESULTS: mismatched parameters and model type")


class VersionWarning(Warning):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy
import json
import sys
import time
import warnings
from datetime import datetime
from importlib.metadata import version
from typing import TYPE_CHECKING

import jwt
import numpy as np
import pandas as pd
import requests
import semver
from bert_schemas import job as job_schema
from pydantic import ValidationError
from requests.exceptions import RequestException
from tabulate import tabulate

from oqtant.fixtures.jobs import barrier_manipulator_job, ultracold_matter_job
from oqtant.schemas.job import OqtantJob
from oqtant.settings import Settings
from oqtant.simulator import QMPotential, Simulator
from oqtant.util import exceptions as api_exceptions
from oqtant.util.auth import get_user_token

if TYPE_CHECKING:
    from oqtant.schemas.quantum_matter import QuantumMatter

settings = Settings()


class OqtantClient:
    """Python class for interacting with Oqtant
    This class contains tools for:
        - Accessing all of the functionality of the Oqtant Web App (https://oqtant.infleqtion.com)
            - BARRIER (Barrier Manipulator) jobs
            - BEC (Ultracold Matter) jobs
        - Building parameterized (i.e. optimization) experiments using QuantumMatter
        - Submitting and retrieve results
    How Oqtant works:
        - Instantiate a QuantumMatterFactory and log in with your Oqtant account
        - Create QuantumMatter objects with the QuantumMatterFactory
            - 1D parameter sweeps are supported
        - Submit the QuantumMatter to Oqtant to be run on the hardware in a FIFO queue
            - Once submitted a job is created and associated with the QuantumMatter object
        - Retrieve the results of the job from Oqtant into the QuantumMatter object
            - These results are available in future python sessions
        - Extract, visualize, and analyze the results
    Need help? Found a bug? Contact oqtant@infleqtion.com for support. Thank you!
    """

    def __init__(self, *, settings, token, debug: bool = False):
        self.token: str = token
        self.max_ind_var: int = settings.max_ind_var
        self.run_list_limit: int = settings.run_list_limit
        self.debug: bool = debug
        self.version = version("oqtant")
        self.verbosity: int = 1

        self.base_url: str = f"{settings.base_url}/users/{self.external_user_id}"

        if not self.debug:
            sys.tracebacklimit = 0

    @property
    def external_user_id(self) -> str:
        try:
            token_data = jwt.decode(self.token, key=None, options={"verify_signature": False})
            external_user_id = token_data["sub"].replace("auth0|", "")
        except Exception as err:
            raise api_exceptions.OqtantTokenError(
                "Unable to decode JWT token. Please confirm that your token is correct."
            ) from err
        return external_user_id

    def __get_headers(self) -> dict:
        """Method to generate headers for use in calls to the REST API with requests

        Returns:
            dict: A dict of header information
        """
        return {
            "Authorization": f"Bearer {self.token}",
            "X-Client-Version": version("oqtant"),
        }

    def __print(self, message: str) -> None:
        """Method to control the verbosity of the print statements

        Args:
            message (str): The message to print
        """
        if self.verbosity >= 1:
            print(message)

    def convert_matter_to_job(self, matter: QuantumMatter) -> OqtantJob | None:
        """Method to convert a QuantumMatter object to an OqtantJob object

        Args:
            matter (oqtant.schemas.quantum_matter.QuantumMatter): The QuantumMatter object to be converted

        Returns:
            oqtant.schemas.job.OqtantJob: The resulting OqtantJob object
        """

        if matter.result:
            return matter.result

        if matter.output:
            inputs = {
                "values": matter.input.model_dump(),
                "notes": matter.note,
                "output": {"values": matter.output.model_dump()},
            }
        else:
            inputs = {"values": matter.input.model_dump(), "notes": matter.note}

        job_data = {
            "name": matter.name,
            "inputs": [job_schema.Input(**inputs)],
        }

        return OqtantJob(**job_data)

    def get_sim_result(self, sim: Simulator, image_type: job_schema.ImageType) -> Simulator:
        """Method to get the result of a simulation

        Args:
            sim (oqtant.simulator.Simulator): The Simulator object

        Returns:
            oqtant.simulator.Simulator: The Simulator object
        """
        print("ground state calculation in progress       ", end="\n")
        sim.set_ground_state()
        print("ground state calculation complete          ")
        print("trapped simulation in progress             ", end="\n")
        sim.run_evolution()
        print("trapped simulation complete                ")

        if image_type == job_schema.ImageType.TIME_OF_FLIGHT:
            print("time-of-flight simulation in progress              ")
            sim.run_TOF()

        return sim

    def submit_sim(self, matter: QuantumMatter) -> Simulator:
        """Method to submit a QuantumMatter object for simulation

        Args:
            matter (oqtant.schemas.quantum_matter.QuantumMatter): A QuantumMatter object

        Returns:
            oqtant.simulator.Simulator: The Simulator object
        """
        potential = QMPotential(matter)
        sim = Simulator(potential)  # wavefunction, potential).  wavefunction is no longer needed to be passed.
        return sim

    def submit(self, matter: QuantumMatter, track: bool = False, sim: bool = False) -> str:
        """Method to submit a QuantumMatter object for execution, returns the resulting job id

        Args:
            matter (oqtant.schemas.quantum_matter.QuantumMatter): The QuantumMatter object to submit for execution
            track (bool, optional): Flag to track the status of the resulting job
            sim (bool, optional): Flag to submit job as a simulation

        Returns:
            str: The Job ID of the submitted job
        """
        matter.result = None
        if sim:
            self.submit_sim(matter)

        job = self.convert_matter_to_job(matter)
        return self.run_jobs(job_list=[job], track_status=track)[0]

    def submit_list_as_batch(
        self,
        matter_list: list[QuantumMatter],
        track: bool = False,
        name: str | None = None,  # optional global name for resulting job
        sim: bool = False,
    ) -> QuantumMatter:
        """Method to submit a list of QuantumMatter objects as a batch job for execution

        Args:
            matter_list (list[oqtant.schemas.quantum_matter.QuantumMatter]): The list of QuantumMatter objects to
                submit as a single batch job
            track (bool, optional): Whether to track the status of the job
            name (str | None, optional): The name of the batch job. If None, the name of the first program will be used
            sim (bool): If the user intended to submit a sim job as batch or not. Will throw an exception as we
            do not allow simulator jobs as batch.
        Returns:
            str: The ID of the submitted job
        """

        if sim:
            raise api_exceptions.SimSubmitError

        mattersmost = matter_list[0]
        if name:
            mattersmost.name = name

        inputs = []
        master_job_type = self.convert_matter_to_job(mattersmost).job_type
        for matter in matter_list:
            job_type = self.convert_matter_to_job(matter).job_type
            if job_type is not master_job_type:
                raise api_exceptions.OqtantError("All input objects must map to the same job type.")
            inputs.append({"values": matter.input.model_dump(), "notes": matter.note})
        job_data = {
            "name": name,
            "inputs": inputs,
        }
        job = OqtantJob(**job_data)
        id = self.run_jobs(job_list=[job], track_status=track)[0]
        mattersmost.job_id = id
        return mattersmost

    def get_job(self, job_id: str, run: int = 1) -> OqtantJob:
        """Method to get an OqtantJob from the Oqtant REST API. This will always be a targeted query
        for a specific run. If the run is omitted then this will always return the first run of the
        job. Will return results for any job regardless of it's status

        Args:
            job_id (str): This is the external_id of the job to fetch
            run (int, optional): The run to target, this defaults to the first run if omitted

        Returns:
            oqtant.schemas.job.OqtantJob: An OqtantJob instance with the values of the job queried
        """
        request_url = f"{self.base_url}/jobs/{job_id}"
        params = {"run": run}
        response = requests.get(
            url=request_url,
            params=params,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.OqtantAuthorizationError
        try:
            response.raise_for_status()
        except RequestException as err:
            raise api_exceptions.OqtantRequestError(f"Failed to get job '{job_id}' from Oqtant") from err
        job_data = response.json()
        try:
            job = OqtantJob(**job_data)
        except ValidationError as err:
            raise api_exceptions.OqtantJobValidationError(f"Failed to validate job '{job_id}'") from err
        except (KeyError, Exception) as err:
            raise api_exceptions.OqtantJobParameterError(f"Failed to parse job '{job_id}'") from err
        return job

    # necessary?
    def generate_oqtant_job(self, *, job: dict) -> OqtantJob:
        """Method to generate an instance of OqtantJob from the provided dictionary that contains the job
        details and input. Will validate the values and raise an informative error if any
        violations are found

        Args:
            job (dict): Dictionary containing job details and input

        Returns:
            oqtant.schemas.job.OqtantJob: an OqtantJob instance containing the details and input from the
                provided dictionary
        """
        try:
            oqtant_job = OqtantJob(**job)
        except (KeyError, ValidationError) as err:
            raise api_exceptions.OqtantJobValidationError("Failed to generate OqtantJob") from err
        return oqtant_job

    def create_job(
        self,
        name: str,
        job_type: job_schema.JobType,
        runs: int = 1,
        job: dict | None = None,
    ) -> OqtantJob:
        """Method to create an instance of OqtantJob. When not providing a dictionary of job data this method will
        return an OqtantJob instance containing predefined input data based on the value of job_type and runs.
        If a dictionary is provided an OqtantJob instance will be created using the data contained within it.

        Args:
            name (str): The name of the job to be created
            job_type (bert_schemas.job.JobType): The type of job to be created
            runs (int): The number of runs to include in the job
            job (dict | None, optional): Dictionary of job inputs to use instead of the defaults

        Returns:
            oqtant.schemas.job.OqtantJob: an OqtantJob instance of the provided dictionary or predefined input data
        """
        if job:
            job["name"] = name
            job["job_type"] = job_type
            job = self.generate_oqtant_job(job=job)
            return job
        if job_type == job_schema.JobType.BARRIER:
            job = barrier_manipulator_job
        elif job_type == job_schema.JobType.BEC:
            job = ultracold_matter_job
        else:
            raise api_exceptions.OqtantJobUnsupportedTypeError(
                f"Job type '{job_type}' is either invalid or unsupported"
            )
        job = self.generate_oqtant_job(job=job)
        job.inputs = [copy.deepcopy(job.inputs[0].model_dump())] * runs
        job.name = name
        return job

    def submit_job(self, *, job: OqtantJob, write: bool = False) -> dict:
        """Method to submit a single OqtantJob to the Oqtant REST API. Upon successful submission this method will
        return a dictionary containing the external_id of the job and it's position in the queue. Will write the job
        data to file when the write argument is True.



        Args:
            job (oqtant.schemas.job.OqtantJob): The OqtantJob instance to submit for processing
            write (bool, optional): Flag to write job data to file

        Returns:
            dict: Dictionary containing the external_id of the job and it's queue position
        """
        if not isinstance(job, OqtantJob):
            try:
                job = OqtantJob(**job)
            except (TypeError, AttributeError, ValidationError) as err:
                raise api_exceptions.OqtantJobValidationError("OqtantJob is invalid") from err
        data = {
            "name": job.name,
            "job_type": job.job_type,
            "input_count": len(job.inputs),
            "inputs": [input.model_dump() for input in job.inputs],
        }
        response = requests.post(
            url=f"{self.base_url}/jobs",
            json=data,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.OqtantAuthorizationError
        try:
            response.raise_for_status()
        except RequestException as err:
            raise api_exceptions.OqtantRequestError("Failed to submit job to Oqtant") from err
        response_data = response.json()
        if write:
            job.status = job_schema.JobStatus.PENDING
            job.external_id = response_data["job_id"]
            self.write_job_to_file(job)
        return response_data

    def cancel_job(self, job_id: str) -> None:
        """Method to cancel a single job with the Oqtant REST API

        Args:
           job_id (str): The job id of the job to cancel
        """
        response = requests.put(
            url=f"{self.base_url}/jobs/{job_id}/cancel",
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.OqtantAuthorizationError
        try:
            response.raise_for_status()
        except RequestException as err:
            raise api_exceptions.OqtantRequestError("Failed to cancel job") from err

    def delete_job(self, job_id: str) -> None:
        """Method to delete a single job with the Oqtant REST API

        Args:
           job_id (str): The job id of the job to delete
        """
        response = requests.delete(
            url=f"{self.base_url}/jobs/{job_id}",
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.OqtantAuthorizationError
        try:
            response.raise_for_status()
        except RequestException as err:
            raise api_exceptions.OqtantRequestError("Failed to delete job") from err

    def run_jobs(self, job_list: list[OqtantJob], track_status: bool = False, write: bool = False) -> list[str]:
        """Method to submit a list of OqtantJobs to the Oqtant REST API. This method provides some optional
        functionality to alter how it behaves. Providing it with an argument of track_status=True will make it wait
        and poll the Oqtant REST API until all jobs in the list have completed. Providing it with and argument of
        write=True  will make it write the results of the jobs to file when they complete (only applies when the
        track_status argument is True)

        Args:
            job_list (list[oqtant.schemas.job.OqtantJob]): The list of OqtantJob instances to submit for processing
            track_status (bool, optional): Flag to return immediately or wait and poll until all jobs have completed
            write (bool, optional): Flag to write job results to file

        Returns:
            list[str]: List of the external_id(s) returned for each submitted job in job_list
        """
        if len(job_list) > self.run_list_limit:
            raise api_exceptions.OqtantJobListLimitError(
                f"Maximum number of jobs submitted per run is {self.run_list_limit}."
            )
        pending_jobs = []
        submitted_jobs = []
        self.__print(f"Submitting {len(job_list)} job(s):")
        for job in job_list:
            response = self.submit_job(job=job, write=write)
            external_id = response["job_id"]
            job.external_id = external_id
            pending_jobs.append(job)
            submitted_jobs.append(job)
            self.__print(f"\n- Job: {job.name}")
            self.__print(f"  Job ID: {job.external_id}")
        if track_status:
            self.track_jobs(pending_jobs=pending_jobs, write=write)
        return [str(job.external_id) for job in submitted_jobs]

    def search_jobs(
        self,
        *,
        job_type: job_schema.JobType | None = None,
        name: job_schema.JobName | None = None,
        submit_start: str | None = None,
        submit_end: str | None = None,
        notes: str | None = None,
        limit: int = 100,
        show_results: bool = False,
    ) -> list[dict]:
        """Method to submit a query to the Oqtant REST API to search for jobs that match the provided criteria.
        The search results will be limited to jobs that meet your Oqtant account access

        Args:
           job_type (bert_schemas.job.JobType | None, optional): The type of the jobs to search for
           name (bert_schemas.job.JobName | None, optional): The name of the job to search for
           submit_start (str | None, optional): The earliest submit date of the jobs to search for
           submit_start (str | None, optional): The latest submit date of the jobs to search for
           notes (str | None, optional): The notes of the jobs to search for
           limit (int, optional): The limit for the number of jobs returned (max: 100)
           show_results (bool, optional): Flag to print out the results of the search

        Returns:
           list[dict]: A list of jobs matching the provided search criteria
        """
        params = {"limit": limit}
        for param in ["job_type", "name", "submit_start", "submit_end", "notes"]:
            if locals()[param] is not None:
                params[param] = locals()[param]

        response = requests.get(
            url=f"{self.base_url}/jobs",
            params=params,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.OqtantAuthorizationError
        try:
            response.raise_for_status()
        except RequestException as err:
            raise api_exceptions.OqtantRequestError("Failed to search jobs in Oqtant") from err
        job_data = response.json().get("items", [])
        if show_results and job_data:
            self.__print(f"Search returned {len(job_data)} job(s):\n")
            rows = [
                [
                    (job.get("name") if len(job.get("name", [])) < 40 else f"{job.get('name')[:37]}..."),
                    job.get("job_type"),
                    job.get("status"),
                    job.get("external_id"),
                ]
                for job in job_data
            ]
            table = [
                ["Name", "Job Type", "Status", "ID"],
                *rows,
            ]
            self.__print(tabulate(table, headers="firstrow", tablefmt="fancy_grid"))
        return job_data

    def track_jobs(self, pending_jobs: list[OqtantJob], write: bool = False) -> None:
        """Method that polls the Oqtant REST API with a list of OqtantJobs and waits until all of them have
        completed. Will output each job's status while it is polling and will output a message when
        all jobs have completed. When the write argument is True it will also write the jobs' data
        to file when they complete

        Args:
           pending_jobs (list[oqtant.schemas.job.OqtantJob]): List of OqtantJobs to track
           write (bool, optional): Flag to write job results to file
        """
        self.__print(f"\nTracking {len(pending_jobs)} job(s):")
        pending_jobs = {str(job.external_id): job for job in pending_jobs}
        running_job = None
        while pending_jobs:
            if not running_job:
                for external_id, pending_job in pending_jobs.items():
                    job = self.get_job(job_id=external_id)
                    if job.status != pending_job.status:
                        pending_jobs[external_id].status = job.status
                        running_job = pending_jobs[external_id]
                        self.__print(f"\n- Job: {job.name}")
                        self.__print(f"  - {job.status}")
                        break
            else:
                job = self.get_job(job_id=running_job.external_id)
                if job.status != running_job.status:
                    if job.status in [
                        job_schema.JobStatus.INCOMPLETE,
                        job_schema.JobStatus.FAILED,
                        job_schema.JobStatus.COMPLETE,
                    ]:
                        pending_jobs.pop(str(running_job.external_id), None)
                        running_job = None
                    else:
                        running_job.status = job.status
                    self.__print(f"  - {job.status}")
                    if not running_job and write:
                        self.write_job_to_file(job)
            time.sleep(2)
        self.__print("\nAll job(s) complete")

    def write_job_to_file(
        self,
        job: OqtantJob,
        file_name: str | None = None,
        file_path: str | None = None,
    ) -> None:
        """Method to write an OqtantJob instance to a file

        Args:
            job (oqtant.schemas.job.OqtantJob): The OqtantJob instance to write to file
            file_name (str | None, optional): custom name of the file
            file_path (str | None, optional): full path to the file to write, including
                the name of the file
        """
        if file_path:
            target = file_path
        else:
            if job.input_count > 1:
                target = f"{file_name if file_name else str(job.external_id)}"
                target = f"{target}_run_{job.run}_of_{job.input_count}.txt"
            else:
                target = f"{file_name if file_name else str(job.external_id)}.txt"
        try:
            with open(target, "w+") as f:
                f.write(str(job.model_dump_json()))
                print(f'Wrote file: "{target}"')
        except Exception as err:
            raise api_exceptions.JobWriteError(f"Failed to write job to '{target}'") from err

    def load_job_from_file(self, file_path: str, refresh: bool = True) -> OqtantJob:
        """Method to load an OqtantJob instance from a file. Will refresh the job data from the
        Oqtant REST API by default

        Args:
            file_path (str): The full path to the file to read
            refresh (bool, optional): Flag to refresh the job data from Oqtant

        Returns:
            OqtantJob: An OqtantJob instance of the loaded job
        """
        try:
            with open(file_path) as json_file:
                data = json.load(json_file)
                job = OqtantJob(**data)
                run = job.inputs[0].run
                if refresh:
                    job = self.get_job(job.external_id, run=run)
                    self.write_job_to_file(job, file_path=file_path)
                return job
        except FileNotFoundError as err:
            raise api_exceptions.JobReadError(f"Failed to load job from {file_path}") from err
        except (ValidationError, KeyError) as err:
            raise api_exceptions.JobReadError(f"Failed to parse job from {file_path}") from err
        except RequestException as err:
            raise api_exceptions.JobReadError(f"Failed to refresh job from {file_path}") from err

    def get_raw_images(self, job_id: str) -> list:
        """Method to get raw images from the Oqtant REST API

        Args:
            job_id (str): The external_id of the job to fetch

        Returns:
            dict: Dictionary of raw images
        """
        url = f"{self.base_url}/jobs/{job_id}/raw_images"
        response = requests.get(
            url=url,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.OqtantAuthorizationError
        try:
            response.raise_for_status()
        except RequestException as err:
            raise api_exceptions.OqtantRequestError("Failed to get raw images from Oqtant") from err
        raw_images = response.json()
        return raw_images

    def get_job_limits(self, show_results: bool = False) -> dict:
        """Method to get job limits from the Oqtant REST API

        Args:
            show_results (bool, optional): Flag to print out the results

        Returns:
            dict: Dictionary of job limits
        """
        url = f"{self.base_url}/job_limits"
        response = requests.get(
            url=url,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.OqtantAuthorizationError()
        try:
            response.raise_for_status()
        except RequestException as err:
            raise api_exceptions.OqtantRequestError("Failed to get job limits from Oqtant") from err
        job_limits = response.json()
        if show_results:
            limit_table = [
                [
                    "Daily Limit",
                    "Daily Remaining",
                    "Standard Credits",
                    "Priority Credits",
                ],
                [
                    job_limits["daily_limit"],
                    job_limits["daily_remaining"],
                    job_limits["standard_credits"],
                    job_limits["priority_credits"],
                ],
            ]
            self.__print("Job Limits:\n" + tabulate(limit_table, headers="firstrow", tablefmt="fancy_grid"))
        return job_limits

    def show_job_limits(self) -> None:
        """Method to print out job limit information
        Wrapper for OqtantClient.get_job_results
        """
        self.get_job_limits(show_results=True)

    def get_queue_status(
        self,
        job_type: job_schema.JobType | None = None,
        name: job_schema.JobName | None = None,
        submit_start: str | None = None,
        submit_end: str | None = None,
        note: str | None = None,
        limit: int = 50,
        include_complete: bool = False,
        show_results: bool = False,
    ) -> list:
        """Method to get the queue status of jobs submitted by the authenticated user

        Args:
            job_type (bert_schemas.job.JobType | None, optional): The type of jobs to filter results on
            name (bert_schemas.job.JobName | None, optional): The name of the job(s) to filter results on
            submit_start (str, optional): The earliest job submission date to filter results on
            submit_end (str, optional): The latest job submission date to filter results on
            note (str | None, optional): The notes value to filter results on
            limit (int, optional): The limit on the number of results to be returned
            include_complete (bool, optional): Flag to include completed jobs in results
            show_results (bool, optional): Flag to print out the results

        Returns:
            list[dict]: List of jobs that matched the provided query parameters
        """
        now = datetime.now().isoformat()
        start = submit_start if submit_start else now
        end = submit_end if submit_end else now
        search_results = list(
            filter(
                lambda job: (True if include_complete else job.get("status") != job_schema.JobStatus.COMPLETE),
                self.search_jobs(
                    job_type=job_type,
                    name=name,
                    submit_start=start,
                    submit_end=end,
                    notes=note,
                    limit=limit,
                ),
            )
        )
        if show_results:
            self.__print(f"{len(search_results)} job(s) queued:\n")
            rows = [
                [
                    (job.get("name") if len(job.get("name", [])) < 40 else f"{job.get('name')[:37]}..."),
                    job.get("status"),
                    OqtantJob.format_datetime(job.get("time_submit")),
                    job.get("external_id"),
                ]
                for job in search_results
            ]
            table = [
                ["Name", "Status", "Submit", "ID"],
                *rows,
            ]
            self.__print(tabulate(table, headers="firstrow", tablefmt="fancy_grid"))
        return search_results

    def show_queue_status(self, *args, **kwargs) -> None:
        """Method to show queue status information
        Wrapper for OqtantClient.get_queue_status
        """
        self.get_queue_status(*args, **kwargs, show_results=True)

    def check_version(self) -> bool:
        """Method to compare the currently installed version of Oqtant with the latest version in PyPi
        and will raise a warning if it is older

        Returns:
            bool: True if current version is latest, False if it is older
        """
        resp = requests.get("https://pypi.org/pypi/oqtant/json", timeout=5)
        current = True
        if resp.status_code == 200:
            current_version = resp.json()["info"]["version"]
            if semver.compare(self.version, current_version) < 0:
                current = False
                warnings.warn(
                    f"Please upgrade to Oqtant version {current_version}. "
                    f"You are currently using version {self.version}"
                )
        return current


def get_oqtant_client(token: str) -> OqtantClient:
    """Method to create a new OqtantClient instance.

    Args:
        token (str): The auth0 token required for interacting with the Oqtant REST API

    Returns:
        OqtantClient: Authenticated instance of OqtantClient
    """
    client = OqtantClient(settings=settings, token=token)
    client.check_version()
    client.show_job_limits()
    return client


def get_client(port: int = 8080) -> OqtantClient:
    """Method to get both an authentication token and an instance of OqtantClient

    Args:
        port (int, optional): Specific port to run the authentication server on

    Returns:
        OqtantClient: An authenticated instance of OqtantClient
    """
    token = get_user_token(auth_server_port=port)
    client = OqtantClient(settings=settings, token=token)
    return client
